#!/usr/bin/env python
# encoding: utf-8

from typing import Dict, Set, Optional, Iterator, Iterable

from .utils import (
    safe_divide,
    safe_dist,
    get_independent_bin,
    safe_random_choice,
    safe_random_int,
)
from .location import Location
from . import features
from . import exposures


class Agent:
    """
    This class constructs and represents an agent within the population
    """

    # class variable for agent creation
    next_agent_id = 0

    @classmethod
    def update_id_counter(cls, last_id):
        cls.next_agent_id = last_id + 1

    def __init__(
        self,
        sex_type: str,
        age: int,
        race: str,
        drug_use: str,
        location: Location,
        id: Optional[int] = None,
    ) -> None:
        """
        Initialize an agent based on given properties

        Args:
            id: Unique agent ID
            sex_type: Name of defined sex type (e.g. MSM) [params.classes.sex_types]
            age: Agents initialization age
            race: Race of agent [params.classes.races]
            drug_use: Drug use flag [params.classes.drug_types]
        """
        # self.id is unique ID number used to track each person agent.
        if id is not None:
            self.id = id
        else:
            self.id = self.next_agent_id

        self.update_id_counter(self.id)

        # agent properties
        self.sex_type = sex_type
        self.age = age
        self.age_bin = 0
        self.race = race
        self.drug_type = drug_use
        self.location = location
        self.component = "-1"  # updated after relationships created

        self.sex_role = "versatile"

        # agent-partner params
        self.relationships: Set[Relationship] = set()
        self.partners: Dict[str, Set] = {}
        self.mean_num_partners: Dict[str, int] = {}
        self.target_partners: Dict[str, int] = {}

        # agent exposures params
        # model features
        for exposure in exposures.BaseExposure.__subclasses__():
            setattr(self, exposure.name, exposure(self))

        # model features
        for feature in features.BaseFeature.__subclasses__():
            setattr(self, feature.name, feature(self))

    def __str__(self) -> str:
        """
        String formatting of agent object

        returns:
            String formatted tab-deliminated agent properties
        """
        return (
            f"\t{self.id}\t{self.age}\t{self.sex_type}\t{self.drug_type}\t"  # type: ignore[attr-defined]
            f"{self.race}\t{self.hiv.active}"  # type: ignore[attr-defined]
        )

    def __repr__(self) -> str:
        """
        Repr formatting of agent object

        returns:
            agent ID as str
        """
        return str(self.id)

    def __eq__(self, other) -> bool:
        return self.id == other.id

    def __ne__(self, other) -> bool:
        return self.id != other.id

    def __hash__(self) -> int:
        return self.id

    def iter_partners(self) -> Iterator["Agent"]:
        """
        Get an iterator over an agent's partners

        returns:
            iterator of agent partners
        """
        for partner_set in self.partners.values():
            for partner in partner_set:
                yield partner

    def has_partners(self) -> bool:
        """
        Determine whether an agent has any partners

        returns:
            whether an agent has at least one partner
        """
        return any(self.iter_partners())

    def is_msm(self) -> bool:
        """
        Determine whether an agent is a man who can have sex with men

        returns:
            if agent is MSM
        """
        sex_dict = self.location.params.classes.sex_types
        if sex_dict[self.sex_type].gender != "M":
            return False

        for sex_type in sex_dict[self.sex_type].sleeps_with:
            if sex_dict[sex_type].gender == "M":
                return True
        return False

    def get_partners(self, bond_types: Optional[Iterable[str]] = None) -> Set["Agent"]:
        """
        Get all of an agents partners or those with specific bond types

        args:
            bond_types: list of bond types which will filter the partners, otherwise all partners returned

        returns:
            set of agent's partners
        """
        if bond_types:
            partners = set()
            for bond in bond_types:
                partners.update(self.partners[bond])
        else:
            partners = {partner for partner in self.iter_partners()}

        return partners

    def get_num_partners(self, bond_types: Optional[Iterable[str]] = None) -> int:
        """
        Get the number of partners an agent has, optionally filtered by bond type

        args:
            bond_types: list of bond types which will filter the partners, otherwise total number of partners returned

        returns:
            the number of partners the agent has
        """
        return len(self.get_partners(bond_types))


class Relationship:
    """Class for agent relationships."""

    # class variable for relationship creation
    next_rel_id = 0

    @classmethod
    def update_id_counter(cls, last_id):
        cls.next_rel_id = last_id + 1

    def __init__(
        self,
        agent1: Agent,
        agent2: Agent,
        duration: int,
        bond_type: str,
        id: Optional[int] = None,
    ):
        """
        Constructor for a Relationship

        args:
            agent1: first agent
            agent2: second agent
            duration: target duration of relationship
            bond_type: type of bond for the relationship [params.classes.bond_types]
            id: unique identifier
        """
        # make sure these agents can be in a relationship
        assert agent1 != agent2, "Cannot create relationship with same agent"
        for rel in agent1.relationships:
            assert agent2 != rel.get_partner(agent1), "Agents already partnered!"

        # self.id is unique ID number used to track each person agent.
        self.agent1 = agent1
        self.agent2 = agent2

        if id is not None:
            self.id = id
        else:
            self.id = self.next_rel_id

        self.update_id_counter(self.id)

        # Relationship properties
        self.duration = duration
        self.total_duration = duration
        self.bond_type = bond_type

        self.bond()

    def __eq__(self, other) -> bool:
        return self.id == other.id

    def __ne__(self, other) -> bool:
        return self.id != other.id

    def __hash__(self) -> int:
        return self.id

    def progress(self, force: bool = False) -> bool:
        """
        Progress a relationship to the next time step (decrementing remaining target duration), or end a relationship if the duration is 0 or if `force` is set to `True`

        args:
            force: whether to force the relationship to end
        """
        if self.duration <= 0 or force:
            self.unbond()
            return True
        else:
            self.duration -= 1
            return False

    def bond(self) -> None:
        """
        Bond two agents. Adds the relationship to each agent's `relationships` set, then adds each partner to the others' partner list.
        """

        # Append relationship to relationships list for each agent
        self.agent1.relationships.add(self)
        self.agent2.relationships.add(self)

        # Pair agent with partner and partner with agent
        self.agent1.partners[self.bond_type].add(self.agent2)
        self.agent2.partners[self.bond_type].add(self.agent1)

    def unbond(self):
        """
        Unbond two agents. Removes relationship from relationship sets.
        Removes partners in each others' partner list.
        """

        # Remove relationship to relationships list for each agent
        self.agent1.relationships.remove(self)
        self.agent2.relationships.remove(self)

        # Unpair agent with partner and partner with agent
        self.agent1.partners[self.bond_type].remove(self.agent2)
        self.agent2.partners[self.bond_type].remove(self.agent1)

    def get_partner(self, agent: "Agent") -> "Agent":
        """
        Given an agent in the relationship, return the other agent

        args:
            agent: one of the agents in the relationship

        returns:
            the agent's partner
        """
        if agent == self.agent1:
            return self.agent2
        elif agent == self.agent2:
            return self.agent1
        else:
            raise ValueError("Agent must be in this relationship")

    def get_number_of_sex_acts(self, rand_gen) -> int:
        """
        Number of sex acts in the relationship during the time step.

        args:
            rand_gen: np random number generator (e.g. self.run_random in model)

        returns:
            number of sex acts
        """
        agent = safe_random_choice([self.agent1, self.agent2], rand_gen)
        freq_params = agent.location.params.partnership.sex.frequency[self.bond_type]

        if freq_params.type == "bins":
            i = get_independent_bin(rand_gen, freq_params.bins)
            return safe_random_int(
                freq_params.bins[i].min, freq_params.bins[i].max, rand_gen
            )

        elif freq_params.type == "distribution":
            return round(safe_dist(freq_params.distribution, rand_gen))

        else:
            raise Exception("Sex acts must be defined as bin or distribution")

    def __str__(self):
        return (
            f"\t{self.id}\t{self.agent1.id}\t{self.agent2.id}\t{self.duration}\t"
            f"{self.bond_type} "
        )

    def __repr__(self):
        return str(self.id)


class AgentSet:
    """
    Container for agents into heirarchical sets (e.g. all_agents > hiv_agents)
    """

    def __init__(
        self,
        id: str,
        parent: Optional["AgentSet"] = None,
    ):
        """
        Constructor of an AgentSet

        args:
            id: name of the set
            parent: the set this set is a subset of
        """
        # members stores agent set members in a dictionary keyed by ID
        self.id = id
        self.members: Set[Agent] = set()
        self.subset: Dict[str, AgentSet] = {}

        # parent_set stores the parent set if this set is a member of an
        # AgentSet class instance. For example, for a set that is a
        # member of a larger set, the _parent_set for that set  would
        # be that larger set.
        self.parent_set = parent
        if parent:
            parent.add_subset(self)

    def __repr__(self):
        return self.id

    def __str__(self):
        return self.id

    def clear_set(self):
        """
        Clears a set of any members and subsets
        """
        self.members: Set[Agent] = set()
        self.subset: Dict[str, str] = {}

    def __iter__(self) -> Iterator[Agent]:
        """
        Iterate over the memers in the set

        example:
            ```py
            for agent in agent_set:
                # do something with agent
            ```

        returns:
            iterator over member agents
        """
        return self.members.__iter__()

    def __contains__(self, item) -> bool:
        """
        Is an agent a member of this agent set

        example:
            ```py
            if agent in agent_set:
                # do something
            ```

        returns:
            whether agent is part of set
        """
        return self.members.__contains__(item)

    # adding trickles up
    def add_agent(self, agent: Agent) -> None:
        """
        Adds an agent to the set and any parent sets

        args:
            agent: agent to add
        """
        self.members.add(agent)

        if self.parent_set is not None:
            self.parent_set.add_agent(agent)

    # removing trickles down
    def remove_agent(self, agent: Agent) -> None:
        """
        Removes agent from agent set if they are a member of the set.  Also removes the agent from any subsets.

        args:
            agent: agent to remove
        """
        if agent in self.members:
            self.members.remove(agent)

        for subset in self.iter_subset():
            subset.remove_agent(agent)

    def num_members(self) -> int:
        """
        Number of members in the set

        returns:
            number of members
        """
        return len(self.members)

    def add_subset(self, subset: "AgentSet") -> None:
        """
        Adds a new AgentSet to the current sets subset.

        args:
            subset: subset to add to this set
        """
        if subset.id not in self.subset:
            self.subset[subset.id] = subset

    def iter_subset(self) -> Iterator["AgentSet"]:
        """
        Iterate over the subsets of this agent set

        returns:
            iterator of agent sets
        """
        for subset in list(self.subset.values()):
            yield subset

    def print_subsets(self, printer=print):
        """
        Pretty print the subsets of this agent set
        """
        lines = []

        lines.append(f"\t__________ {self.id} __________")
        # lines.append("\tID\t\tN\t\t%")
        lines.append("\t{:^6}\t\t{:^5}\t\t{:^4}".format("ID", "N", "%"))
        for set in self.iter_subset():
            lines.append(
                "\t{:^6}\t\t{:^5}\t\t{:.2}".format(
                    set.id,
                    set.num_members(),
                    safe_divide(set.num_members(), set.parent_set.num_members()),
                )
            )
            for subset in set.iter_subset():
                lines.append(
                    "\t{:4}\t\t{:5}\t\t{:.2}".format(
                        subset.id,
                        subset.num_members(),
                        safe_divide(
                            subset.num_members(), subset.parent_set.num_members()
                        ),
                    )
                )
        lines.append("\t______________ END ______________")
        printer("\n".join(lines))



from typing import Dict, Optional

import numpy as np  # type: ignore

from . import base_feature
from .. import agent
from .. import population
from .. import model


class Vaccine(base_feature.BaseFeature):

    name = "vaccine"
    stats = ["vaccine"]
    """
        Vaccine collects the following stats:

        * vaccine - number of agents with active vaccine
    """

    def __init__(self, agent: "agent.Agent"):
        super().__init__(agent)
        self.active = False
        self.time: Optional[int] = None
        self.type = ""

    def init_agent(self, pop: "population.Population", time: int):
        """
        Initialize the agent for this feature during population initialization (`Population.create_agent`).  Called on only features that are enabled per the params.

        If the agent is HIV-, randomly vaccinate per the params.

        args:
            pop: the population this agent is a part of
            time: the current time step
        """
        if (
            not self.agent.hiv.active  # type: ignore[attr-defined]
            and self.agent.location.params.vaccine.on_init
            and pop.pop_random.random()
            < self.agent.location.params.demographics[self.agent.race]
            .sex_type[self.agent.sex_type]
            .vaccine.init
        ):
            self.vaccinate(time)

    def update_agent(self, model: "model.TITAN"):
        """
        Update the agent for this feature for a time step.  Called once per time step in `TITAN.update_all_agents`. Agent level updates are done after population level updates.   Called on only features that are enabled per the params.

        If PrEP feature is enable and the agent is not active PrEP and not HIV, either update or stochastically vaccinate the agent.

        args:
            model: the instance of TITAN currently being run
        """
        if (
            not self.agent.prep.active  # type: ignore[attr-defined]
            and not self.agent.hiv.active  # type: ignore[attr-defined]
        ):
            vaccine_params = self.agent.location.params.vaccine
            agent_params = (
                self.agent.location.params.demographics[self.agent.race]
                .sex_type[self.agent.sex_type]
                .vaccine
            )

            if self.active:
                if (
                    vaccine_params.booster
                    and (model.time - self.time) == agent_params.booster.interval
                    and model.run_random.random() < agent_params.booster.prob
                ):
                    self.vaccinate(model.time)
            elif model.time == vaccine_params.start_time:
                if model.run_random.random() < agent_params.prob:
                    self.vaccinate(model.time)

    def set_stats(self, stats: Dict[str, int], time: int):
        if self.active:
            stats["vaccine"] += 1

    def get_acquisition_risk_multiplier(self, time: int, interaction_type: str):
        """
        Get a multiplier for how vaccine affects acquisition of HIV for the given interaction_type.

        By default, returns 1.0

        args:
            time: the current model time step
            interaction_type: The type of interaction where the agent could acquire HIV (e.g. 'sex', 'injection' - from [params.classes.interaction_types])
        """
        # not protected the time step the agent is vaccinaetd
        if self.active and self.time is not None and self.time < time:
            vaccine_time_months = (
                (time - self.time)
                / self.agent.location.params.model.time.steps_per_year
            ) * 12

            if self.type == "HVTN702":
                return np.exp(
                    -2.88 + 0.76 * (np.log((vaccine_time_months + 0.001) * 30))
                )
            elif self.type == "RV144":
                return np.exp(-2.40 + 0.76 * (np.log(vaccine_time_months)))

        return 1.0

    # ============= HELPER METHODS =============

    def vaccinate(self, time):
        """
        Vaccinate an agent and update relevant fields.
        """
        self.active = True
        self.type = self.agent.location.params.vaccine.type
        self.time = time



from typing import List, Dict

import numpy as np  # type: ignore
import networkx as nx  # type: ignore

from . import base_exposure
from .. import agent as ag
from .. import population
from .. import model
from .. import utils


class Knowledge(base_exposure.BaseExposure):

    name: str = "knowledge"
    stats: List[str] = ["knowledge_aware"]
    """
    Knowledge collects the following stats:

    * knowledge_aware - number of agents with active knowledge
    """

    def __init__(self, agent: "ag.Agent"):
        super().__init__(agent)

        self.active = False
        self.opinion = 0.0

    def init_agent(self, pop: "population.Population", time: int):
        """
        Initialize the agent for this exposure during population initialization (`Population.create_agent`).  Called only on exposures that are enabled per the params.

        Stochastically make agent aware, if aware, set the opinion from the params.

        args:
            pop: the population this agent is a part of
            time: the current time step
        """
        knowledge_params = self.agent.location.params.knowledge
        if pop.pop_random.random() < knowledge_params.init:
            self.active = True

        # Initialize all agents with some opinion, may or may not be active
        self.opinion = utils.get_cumulative_bin(
            pop.pop_random, knowledge_params.opinion.init
        )

    def update_agent(self, model: "model.TITAN"):
        """
        Update the agent for this exposure for a time step.  Called once per time step in `TITAN.update_all_agents`.

        If the knowledge start_time has happened, stochastically convert agents.

        args:
            model: the instance of TITAN currently being run
        """
        knowledge_params = self.agent.location.params.knowledge
        if (
            model.time >= knowledge_params.start_time
            and not self.active
            and model.run_random.random() < knowledge_params.prob
        ):
            self.convert(model)

    def set_stats(self, stats: Dict[str, int], time: int):
        if self.active:
            stats["knowledge_aware"] += 1

    @staticmethod
    def expose(
        model: "model.TITAN",
        interaction: str,
        rel: "ag.Relationship",
        num_acts: int,
    ):
        """
        Expose a relationship to the exposure for a number of acts for a specific interaction type.  Typically, this determines if the exposure can cause conversion/change in one of the agents, then if so, determines the probability of that and then converts the succeptible agent.

        If transmission stochastically occurs, either convert the unaware agent, or if both agents aware, have the higher influence agent influce their partner.

        args:
            model: The running model
            interaction: The type of interaction (e.g. sex, injection)
            rel: The relationship where the interaction is occuring
            num_acts: The number of acts of that interaction
        """
        assert (
            model.params.model.network.enable
        ), "Network must be enabled for knowledge exposure"

        # agent/partner ordering is irrelevant at this point for knowledge transmission
        p = rel.agent1.knowledge.get_transmission_probability(  # type: ignore[attr-defined]
            model, interaction, rel.agent2, num_acts
        )

        if model.run_random.random() < p:
            agent1_aware = rel.agent1.knowledge.active  # type: ignore[attr-defined]
            agent2_aware = rel.agent2.knowledge.active  # type: ignore[attr-defined]

            if agent1_aware and agent2_aware:
                influence(model, rel)
            elif agent1_aware:
                rel.agent2.knowledge.convert(model)  # type: ignore[attr-defined]
            elif agent2_aware:
                rel.agent1.knowledge.convert(model)  # type: ignore[attr-defined]

    def get_transmission_probability(
        self,
        model: "model.TITAN",
        interaction: str,
        partner: "ag.Agent",
        num_acts: int,
    ) -> float:
        """
        Get the probability of knowledge/opinion transmission in this relationship

        args:
            model: The running model
            interaction: The interaction type (e.g. `pca`)
            partner: The agent's partner
            num_acts: The number of interactions the agents had

        returns:
            the probability of knowledge/opinion transmission
        """
        if not interaction == "pca":
            return 0.0

        if self.active and partner.knowledge.active:  # type: ignore[attr-defined]
            p = model.params.knowledge.opinion.prob
        else:
            p = model.params.knowledge.prob

        return utils.total_probability(p, num_acts)

    def convert(self, model: "model.TITAN"):
        """
        Make an agent aware, stochastically make knowledge aware if their opinion meets the threshold.

        args:
            model: The running model
        """
        params = self.agent.location.params.knowledge
        self.active = True  # type: ignore[attr-defined]
        if (
            self.opinion > params.opinion.threshold  # type: ignore[attr-defined]
            and model.run_random.random() < params.feature.prob
        ):
            agent_attr = getattr(self.agent, params.feature.name)
            agent_attr.initiate(model, force=True)


# ===================== HELPER FUNCTIONS ===================
def influence(model: "model.TITAN", rel: "ag.Relationship"):
    """
    The higher influence agent chages the opinion of the other agent to be the mean of their opinions.  If the opinion excedes the threshold, initiate exposure.

    args:
        model: The running model
        rel: a relationship where an agent is influencing their partner
    """

    def get_influence(agent):
        return nx.closeness_centrality(model.pop.graph, agent)

    if get_influence(rel.agent1) > get_influence(rel.agent2):
        agent = rel.agent1
        partner = rel.agent2
    else:
        agent = rel.agent2
        partner = rel.agent1

    partner_init_opinion = partner.knowledge.opinion  # type: ignore[attr-defined]
    partner.knowledge.opinion = np.mean([agent.knowledge.opinion, partner.knowledge.opinion])  # type: ignore[attr-defined]

    # partner crossed threshhold of opinion, stochastically enroll in feature
    params = partner.location.params.knowledge
    if (
        partner_init_opinion
        < params.opinion.threshold
        < partner.knowledge.opinion  # type: ignore[attr-defined]
    ):
        if model.run_random.random() < params.feature.prob:
            agent_attr = getattr(partner, params.feature.name)
            agent_attr.initiate(model, force=True)



from typing import Dict, Optional

from . import base_feature
from .. import agent
from .. import population
from .. import model
from .. import utils


class Incar(base_feature.BaseFeature):

    name = "incar"
    stats = ["incar", "incar_hiv", "new_release", "new_release_hiv"]
    """
        Incar collects the following stats:

        * incar - number of agents with active incar
        * incar_hiv - number of agents with active incar and HIV
        * new_release - number of agents released this timestep
        * new_release_hiv - number of agents releasted this timestep with HIV
    """

    def __init__(self, agent: "agent.Agent"):
        super().__init__(agent)

        self.active = False
        self.time: Optional[int] = None
        self.release_time: Optional[int] = None

    def init_agent(self, pop: "population.Population", time: int):
        """
        Initialize the agent for this feature during population initialization (`Population.create_agent`).  Called on only features that are enabled per the params.

        Run incarceration assignment on an agent.  The duration of incarceration at initialization is different than the ongoing to reflect that agents with longer durations will be more highly represented in that population at any given point in time.

        args:
            pop: the population this agent is a part of
            time: the current time step
        """
        incar_params = (
            self.agent.location.params.demographics[self.agent.race]
            .sex_type[self.agent.sex_type]
            .incar
        )
        jail_duration = incar_params.duration.init

        prob_incar = incar_params.init
        if pop.pop_random.random() < prob_incar:
            self.active = True
            bin = 1
            current_p_value = jail_duration[bin].prob
            p = pop.pop_random.random()

            while p > current_p_value:
                bin += 1
                current_p_value += jail_duration[bin].prob

            self.time = time
            self.release_time = time + pop.pop_random.randrange(
                jail_duration[bin].min, jail_duration[bin].max
            )

    def update_agent(self, model: "model.TITAN"):
        """
        Update the agent for this feature for a time step.  Called once per time step in `TITAN.update_all_agents`. Agent level updates are done after population level updates.   Called on only features that are enabled per the params.

        Incarcerate an agent or update their incarceration variables

        args:
            model: the instance of TITAN currently being run
        """
        hiv_bool = self.agent.hiv.active  # type: ignore[attr-defined]

        if hiv_bool:
            hiv_multiplier = self.agent.location.params.incar.hiv.multiplier
        else:
            hiv_multiplier = 1.0

        # agent is incarcerated
        if self.active:
            # Release agent
            if self.release_time == model.time:
                self.active = False

                # does agent stay on haart
                if hiv_bool:
                    if self.agent.haart.active:  # type: ignore[attr-defined]
                        if (
                            model.run_random.random()
                            <= self.agent.location.params.incar.haart.discontinue
                        ):
                            self.agent.haart.active = False  # type: ignore[attr-defined]
                            self.agent.haart.adherent = False  # type: ignore[attr-defined]

        # should the agent become incarcerated?
        elif model.run_random.random() < (
            self.agent.location.params.demographics[self.agent.race]
            .sex_type[self.agent.sex_type]
            .incar.prob
            * hiv_multiplier
            * model.calibration.incarceration
        ):
            incar_duration = (
                self.agent.location.params.demographics[self.agent.race]
                .sex_type[self.agent.sex_type]
                .incar.duration.prob
            )

            bin = utils.get_cumulative_bin(model.run_random, incar_duration)

            self.time = model.time
            self.release_time = model.time + utils.safe_random_int(
                incar_duration[bin].min, incar_duration[bin].max, model.run_random
            )
            self.active = True

            if hiv_bool:
                if not self.agent.hiv.dx:  # type: ignore[attr-defined]
                    if (
                        model.run_random.random()
                        < self.agent.location.params.incar.hiv.dx
                    ):
                        self.agent.hiv.diagnose(model)  # type: ignore[attr-defined]
                else:  # Then tested and HIV, check to enroll in ART
                    if (
                        model.run_random.random()
                        < self.agent.location.params.incar.haart.prob
                    ):
                        self.agent.haart.adherent = model.run_random.random() < self.agent.location.params.incar.haart.adherence  # type: ignore[attr-defined]
                        # Add agent to HAART class set, update agent params
                        self.agent.haart.active = True  # type: ignore[attr-defined]

    def set_stats(self, stats: Dict[str, int], time: int):
        if self.release_time == time:
            stats["new_release"] += 1
            if self.agent.hiv.active:  # type: ignore[attr-defined]
                stats["new_release_hiv"] += 1

        if self.active:
            stats["incar"] += 1
            if self.agent.hiv.active:  # type: ignore[attr-defined]
                stats["incar_hiv"] += 1

    # ============== HELPER METHODS ================



from . import base_feature
from .. import agent
from .. import model


class PartnerTracing(base_feature.BaseFeature):

    name = "partner_tracing"

    def __init__(self, agent: "agent.Agent"):
        super().__init__(agent)

        self.active = False
        self.time = None

    def update_agent(self, model: "model.TITAN"):
        """
        Update the agent for this feature for a time step.  Called once per time step in `TITAN.update_all_agents`. Agent level updates are done after population level updates.   Called on only features that are enabled per the params.

        If the agent is was diagnosed last time step, trace their partners. If the agent is traced but not diagnosed, stochastically diagnose.  If the agents tracking has expired, mark them as inactive.

        args:
            model: the instance of TITAN currently being run
        """
        params = self.agent.location.params.partner_tracing

        if model.time < params.start_time or model.time > params.stop_time:
            return

        agent_exposure = getattr(self.agent, params.exposure)

        if agent_exposure.active:
            # was this agent diagnosed with the target exposure last time step
            if agent_exposure.dx and agent_exposure.dx_time == model.time - 1:
                for ptnr in self.agent.get_partners(params.bond_type):
                    partner_exposure = getattr(ptnr, params.exposure)
                    if (
                        not partner_exposure.dx
                        and model.run_random.random() < params.prob
                    ):
                        ptnr.partner_tracing.active = True  # type: ignore[attr-defined]
                        ptnr.partner_tracing.time = model.time  # type: ignore[attr-defined]

            # second chance at diagnosis if traced
            if (
                self.active
                and self.time < model.time
                and not agent_exposure.dx
                and model.run_random.random() < params.dx_prob
            ):
                agent_exposure.diagnose(model)

        # stop tracing of this agent if time
        if self.active and model.time >= self.time + params.trace_duration:
            self.active = False
            self.time = None



#!/usr/bin/env python
# encoding: utf-8

from typing import Dict, Any, List, Iterator
import itertools
import os

import networkx as nx  # type: ignore
from numpy import mean  # type: ignore

from .parse_params import ObjMap
from . import utils
from . import agent as ag


def setup_aggregates(params: ObjMap, reportables, classes: List[str]) -> Dict:
    """
    Recursively create a nested dictionary of attribute values to items to count.

    Attributes are classes defined in params, the items counted are:

    * "agents"
    * "deaths"
    * "deaths_hiv"

    Additionally, any feature enabled may have additional stats that are tracked.  See the feature's `stats` attribute.

    args:
        params: model parameters
        classes: which classes to aggregate by [params.outputs.classes]

    returns:
        dictionary of class values to counts
    """
    if classes == []:
        base_stats = {
            "agents": 0,
            "deaths": 0,
            "deaths_hiv": 0,
        }

        for reportable in reportables:
            base_stats.update({stat: 0 for stat in reportable.stats})

        return base_stats

    stats = {}
    clss, *rem_clss = classes  # head, tail
    keys = [k for k in params.classes[clss]]

    for key in keys:
        stats[key] = setup_aggregates(params, reportables, rem_clss)

    return stats


def get_aggregates(params: ObjMap) -> Iterator:
    """
    Get iterator over all attribute combinations for output classes

    args:
        params: model parameters

    returns:
        iterator over attribute combinations
    """
    return itertools.product(
        *[list(k for k in params.classes[clss]) for clss in params.outputs.classes]
    )


def get_agg_val(stats: Dict, attrs: List, key: str) -> int:
    """
    Get the value of a key in stats given the attribute values

    args:
        stats: a nested dictionary of attributes to count
        attrs: a list of attribute values to find the count for
        key: the type of count to get the value of

    returns:
        the count of key for the given attributes
    """
    stats_item = stats
    for attr in attrs:
        stats_item = stats_item[attr]

    return stats_item[key]


def get_stats_item(stats: Dict[str, Any], attrs: List[str], agent: "ag.Agent"):
    """
    Get the leaf node of the stats dictionary for the given attributes and agent.

    args:
        stats: a nested dictionary of attributes to count
        attrs: a list of attribute values to find the count for
        agent: The agent to get the leaf node for

    returns:
        a stats_item dictionary of keys to counts
    """
    stats_item = stats
    for attr in attrs:
        stats_item = stats_item[str(getattr(agent, attr))]

    return stats_item


def add_agent_to_stats(stats_item: Dict[str, int], key: str):
    """
    Update the stats dictionary counts for the key given the agent's attributes

    args:
        stats_item: the leaf node of a nested dictionary of attributes to counts
        key: the type of count to increment
    """
    stats_item[key] += 1


def get_stats(
    all_agents: "ag.AgentSet",
    deaths: List["ag.Agent"],
    params: ObjMap,
    exposures,
    features,
    time: int,
) -> Dict:
    """
    Get the current statistics for a model based on the population, and tracking agent sets from the model.

    args:
        all_agents: all of the agents in the population
        new_hiv.dx: agents who are newly diagnosed with hiv this timestep
        deaths: agents who died this timestep
        params: model parameters

    returns:
        nested dictionary of agent attributes to counts of various items
    """
    reportables = exposures + features
    stats = setup_aggregates(params, reportables, params.outputs.classes)

    # attribute names (non-plural)
    attrs = [clss[:-1] for clss in params.outputs.classes]

    for a in all_agents:
        stats_item = get_stats_item(stats, attrs, a)

        add_agent_to_stats(stats_item, "agents")

        for reportable in reportables:
            agent_feature = getattr(a, reportable.name)
            agent_feature.set_stats(stats_item, time)

    for a in deaths:
        stats_item = get_stats_item(stats, attrs, a)
        add_agent_to_stats(stats_item, "deaths")
        if a.hiv.active:  # type: ignore[attr-defined]
            add_agent_to_stats(stats_item, "deaths_hiv")

    return stats


# ================== Printer Functions =========================
# Each of the following functions takes in the time, seeds, and stats dict for that time
# and prints the appropriate stats to file


def write_report(
    file_name: str,
    run_id: str,
    t: int,
    runseed: int,
    popseed: int,
    stats: Dict,
    params: ObjMap,
    outdir: str,
):
    """
    Core function for writing reports, writes header if file is new, then data based
    on the `params` and `name_map`

    args:
        file_name: Name of the file to write, including the extension (e.g. `MyReport.txt`)
        run_id: unique identifier for this model
        t: current timestep
        runseed: integer used to seed the random number generator for the model
        popseed: integer used to seed the random number generator for the population
        stats: nested dictionary of agent attributes to counts
        params: model parameters
        outdir: path of where to save this file
    """

    def get_stat_names(stats, attrs):
        stat_ref = stats
        for i in range(len(attrs)):
            stat_ref = stat_ref[list(stat_ref.keys())[0]]

        return stat_ref

    f = open(os.path.join(outdir, file_name), "a")
    attrs = [clss[:-1] for clss in params.outputs.classes]
    stat_names = get_stat_names(stats, attrs)

    if f.tell() == 0:
        f.write("run_id\trseed\tpseed\tt\t")  # start header

        # attributes in stats
        f.write("\t".join(attrs))

        # report specific fields
        for name in stat_names:
            f.write(f"\t{name}")

        f.write("\n")

    for agg in get_aggregates(params):
        # don't write row if no agents are in it
        if get_agg_val(stats, agg, "agents") > 0:
            f.write(f"{run_id}\t{runseed}\t{popseed}\t{t}\t")

            f.write("\t".join(agg))  # write attribute values

            for name in stat_names:
                f.write(f"\t{(get_agg_val(stats, agg, name))}")

            f.write("\n")

    f.close()


def basicReport(
    run_id: str,
    t: int,
    runseed: int,
    popseed: int,
    stats: Dict[str, Any],
    params: ObjMap,
    outdir: str,
):
    """
    Standard report writer for basic agent statistics, columns include:

    * "agents": number of agents in the population
    * "deaths": number of agents who died this time period
    * "deaths_hiv": number of agents with HIV who died this time period

    Additionally, any feature enabled may have additional stats that are tracked.  See the feature's `stats` attribute and docs for details.
    """
    write_report("basicReport.txt", run_id, t, runseed, popseed, stats, params, outdir)


# ========================== Other Print Functions =============================


def effective_size(comp):
    total_size = 0
    for node in comp.nodes():
        num_neighbors = len(comp[node])
        if num_neighbors == 0:
            continue

        E = comp.subgraph(comp.neighbors(node))
        total_size += num_neighbors - (2 * E.size()) / num_neighbors

    return total_size


def print_components(
    run_id: str,
    t: int,
    runseed: int,
    popseed: int,
    components: List,
    outdir: str,
):
    """
    Write stats describing the components (sub-graphs) in a graph to file

    args:
        run_id: unique identifer for this run of the model
        t: current timestep
        runseed: integer used to seed the model's random number generator
        popseed: integer used to seed the population's random number generator
        components: a list of graph components
        outdir: path where the file should be saved
        races: the races in the population
    """
    f = open(os.path.join(outdir, f"{run_id}_componentReport_ALL.txt"), "a")

    # if this is a new file, write the header info
    if f.tell() == 0:
        f.write(
            "run_id\trunseed\tpopseed\tt\tcomponent"
            "\tdensity\tEffectiveSize\tdeg_cent\n"
        )

    for (id, comp) in enumerate(components):
        num_nodes = comp.number_of_nodes()

        average_size = effective_size(comp) / num_nodes
        comp_density = nx.density(comp)

        deg_cent = mean(list(nx.degree_centrality(comp).values()))

        f.write(
            f"{run_id}\t{runseed}\t{popseed}\t{t}\t{id}\t"
            f"\t{comp_density:.4f}"
            f"\t{average_size:.4f}\t{deg_cent}\n"
        )

    f.close()


def write_graph_edgelist(graph, path: str, id, time):
    """
    Writes a pipe-delimited edge list to the file `<id>_Edgelist_t<time>.txt`

    args:
        path: directory where the file should be saved
        id: identifier for the network, typically the model's `id`
        time: timestep the edgelist is being written at
    """
    file_path = os.path.join(path, f"{id}_Edgelist_t{time}.txt")
    # Write edgelist with bond type
    nx.write_edgelist(graph, file_path, delimiter="|", data=["type"])


def write_network_stats(graph, path: str, id, time):
    """
    Writes network statistics to the file `<id>_NetworkStats_t<time>.txt`

    args:
        path: directory where the file should be saved
        id: identifier for the network, typically the model's `id`
        time: timestep the edgelist is being written at
    """
    file_path = os.path.join(path, f"{id}_NetworkStats_t{time}.txt")

    components = utils.connected_components(graph)

    outfile = open(file_path, "w")
    outfile.write(nx.info(graph))

    cent_dict = nx.degree_centrality(graph)

    outfile.write(
        "\nNumber of connected components: {}\n".format(
            nx.number_connected_components(graph)
        )
    )

    tot_nodes = 0
    for c in components:
        tot_nodes += c.number_of_nodes()

    outfile.write(
        "Average component size: {}\n".format(
            tot_nodes * 1.0 / nx.number_connected_components(graph)
        )
    )
    outfile.write(
        "Maximum component size: {}\n".format(nx.number_of_nodes(components[0]))
    )
    outfile.write("Degree Histogram: {}\n".format(nx.degree_histogram(graph)))
    outfile.write("Graph density: {}\n".format(nx.density(graph)))
    outfile.write(
        "Average node degree centrality: {}\n".format(
            sum(cent_dict.values()) / len(list(cent_dict.values()))
        )
    )

    outfile.write("Average node clustering: {}\n".format(nx.average_clustering(graph)))
    outfile.close()



from typing import Dict
import logging

from . import base_feature
from .. import utils
from .. import model

import networkx as nx  # type: ignore


class RandomTrial(base_feature.BaseFeature):

    name = "random_trial"
    stats = [
        "random_trial",
        "random_trial_treated",
        "random_trial_suitable",
        "random_trial_treated_hiv",
    ]
    """
        Random Trial collects the following stats:

        * random_trial - number of agents with active random_trial
        * random_trial_treated - number of active agents treated
        * random_trial_treated_hiv - number of HIV+ agents treated
        * random_trial_suitable - number of active agents suitable
    """

    def __init__(self, agent):
        super().__init__(agent)

        self.active = False
        self.treated = False
        self.suitable = False

    @classmethod
    def update_pop(cls, model: "model.TITAN"):
        """
        Update the feature for the entire population (class method).

        Initialize a random trial in the population if time is the random trial start time.

        args:
            model: the instance of TITAN currently being run
        """
        rt_params = model.params.random_trial

        if not model.time == rt_params.start_time:
            return

        assert (
            model.params.model.network.enable
        ), "Network must be enabled for random trial"

        logging.info(f"Starting random trial ({rt_params.choice})")
        components = model.pop.connected_components()

        # set up helper methods based on params
        if rt_params.treatment == "prep":
            assert (
                model.params.features.prep
            ), "Prep feature must be enabled to use the prep random trial treatment"
            treat = treat_prep
            suitable = suitable_prep
        elif rt_params.treatment == "knowledge":
            assert (
                model.params.exposures.knowledge
            ), "Knowledge exposure must be enabled to use the knowledge random trial treatment"
            treat = treat_knowledge
            suitable = suitable_knowledge

        total_nodes = 0
        logging.info(
            f"Number of components {len([1 for comp in components if comp.number_of_nodes()])}",
        )
        for comp in components:
            total_nodes += comp.number_of_nodes()
            if model.run_random.random() < rt_params.prob:
                # Component selected as treatment pod!
                for agent in comp.nodes:
                    agent.random_trial.active = True

                # treat all agents
                if rt_params.choice == "all":
                    for agent in comp.nodes():
                        if suitable(agent, model):
                            treat(agent, model)
                            agent.random_trial.suitable = True
                            agent.random_trial.treated = True

                # chose an agent central to the component
                elif rt_params.choice == "eigenvector":
                    centrality = nx.algorithms.centrality.eigenvector_centrality(comp)
                    assert len(centrality) >= 1, "Empty centrality"
                    ordered_centrality = sorted(centrality, key=centrality.get)

                    # find the most central suitable agent, or if none, use most central
                    intervention_agent = ordered_centrality[0]
                    for agent in ordered_centrality:
                        if suitable(agent, model):
                            intervention_agent = agent
                            intervention_agent.random_trial.suitable = True
                            break

                    intervention_agent.random_trial.treated = True
                    treat(intervention_agent, model)

                # chose an agent that is a bridge in the network
                elif rt_params.choice == "bridge":
                    # list all edges that are bridges
                    all_bridges = list(nx.bridges(comp))
                    suitable_agents = [
                        agent
                        for agents in all_bridges
                        for agent in agents
                        if suitable(agent, model)
                    ]  # all suitable agents in bridges

                    chosen_agent = utils.safe_random_choice(
                        suitable_agents, model.run_random
                    )  # select change agent
                    if chosen_agent is not None:
                        chosen_agent.random_trial.suitable = True  # type: ignore[attr-defined]

                    else:  # if no suitable agents, mark a non-suitable agent
                        chosen_agent = utils.safe_random_choice(
                            list(comp.nodes), model.run_random
                        )

                    chosen_agent.random_trial.treated = True  # type: ignore[attr-defined]
                    treat(chosen_agent, model)

                # chose an agent from the component at random
                elif rt_params.choice == "random":
                    suitable_agents = [
                        agent for agent in comp.nodes if suitable(agent, model)
                    ]

                    # if there are agents who meet eligibility criteria,
                    # select one randomly
                    chosen_agent = utils.safe_random_choice(
                        suitable_agents, model.run_random
                    )

                    if chosen_agent is not None:
                        chosen_agent.random_trial.suitable = True
                    else:  # if no suitable agents, mark a non-suitable agent
                        chosen_agent = utils.safe_random_choice(
                            list(comp.nodes), model.run_random
                        )

                    chosen_agent.random_trial.treated = True  # type: ignore[attr-defined]
                    treat(chosen_agent, model)

        logging.info(f"Total agents in trial: {total_nodes}")

    def set_stats(self, stats: Dict[str, int], time: int):
        if self.active:
            stats["random_trial"] += 1
            if self.treated:
                stats["random_trial_treated"] += 1
                if self.agent.hiv.active:  # type: ignore[attr-defined]
                    stats["random_trial_treated_hiv"] += 1
            if self.suitable:
                stats["random_trial_suitable"] += 1

    # ============= HELPER METHODS ====================


# ============= HELPER FUNCTIONS ==================


def treat_prep(agent, model):
    agent.prep.enroll(model.run_random, model.time)


def suitable_prep(agent, model) -> bool:
    if (
        not agent.hiv.active
        and not agent.prep.active
        and model.run_random.random() < agent.location.params.prep.cap
        and not agent.vaccine.active
    ):
        return True
    else:
        return False


def treat_knowledge(agent, model):
    agent.knowledge.convert(model)


def suitable_knowledge(agent, model) -> bool:
    return not agent.hiv.active



from typing import Dict, ClassVar, Optional

import numpy as np  # type: ignore

from . import base_feature
from .. import agent
from .. import population
from .. import model
from ..parse_params import ObjMap
from .. import exposures


class Prep(base_feature.BaseFeature):

    name = "prep"
    stats = ["prep", "prep_new", "prep_injectable", "prep_oral"]
    """
        PrEP collects the following stats:

        * prep - number of agents with active PrEP
        * prep_new - number of agents who became active PrEP this time step
        * prep_injectable - number of agents on injectable PrEP
        * prep_oral - number of agents on oral PrEP
    """

    # class level attributes to track all Prep agents
    counts: ClassVar[Dict[str, int]] = {}

    def __init__(self, agent: "agent.Agent"):
        super().__init__(agent)
        # agent level attributes
        self.active = False
        self.adherent = False
        self.type = ""
        self.time = None
        self.last_dose_time: Optional[int] = None

    @classmethod
    def init_class(cls, params: "ObjMap"):
        """
        Initialize the counts dictionary for the races in the model.

        args:
            params: the population params
        """
        cls.counts = {race: 0 for race in params.classes.races}

    def init_agent(self, pop: "population.Population", time: int):
        """
        Initialize the agent for this feature during population initialization (`Population.create_agent`).  Called on only features that are enabled per the params.

        If an agent does not have HIV, is PrEP eligible, and time is at least the prep start time, they are randomly asigned to enroll in PrEP.

        args:
            pop: the population this agent is a part of
            time: the current time step
        """
        params = self.agent.location.params
        if self.eligible(time):
            if "Racial" in params.prep.target_model:
                if (
                    pop.pop_random.random()
                    < params.demographics[self.agent.race]
                    .sex_type[self.agent.sex_type]
                    .prep.init
                ):
                    self.enroll(pop.pop_random, time)
            elif pop.pop_random.random() < params.prep.init:
                self.enroll(pop.pop_random, time)

    def update_agent(self, model: "model.TITAN"):
        """
        Update the agent for this feature for a time step.  Called once per time step in `TITAN.update_all_agents`. Agent level updates are done after population level updates.   Called on only features that are enabled per the params.

        If the agent is not hiv and time is at least the prep start time, if the agent is already on PrEP update their PrEP attributes, if the agent isn't on PrEP and is eleigible, initiate PrEP.

        args:
            model: the instance of TITAN currently being run
        """
        if (
            not self.agent.hiv.active  # type: ignore[attr-defined]
            and model.time >= self.agent.location.params.prep.start_time
        ):
            if self.active:
                self.progress(model)
            elif self.eligible(model.time):
                self.initiate(model)

    @classmethod
    def add_agent(cls, agent: "agent.Agent"):
        """
        Add an agent to the class (not instance).

        Add agent to the PrEP counts by race, and add the agent to the set of new agents.

        args:
            agent: the agent to add to the class attributes
        """
        # set up if this is the first time being called
        cls.counts[agent.race] += 1

    @classmethod
    def remove_agent(cls, agent):
        """
        Remove an agent from the class (not instance).

        Decrement the prep counts by race.

        args:
            agent: the agent to remove from the class attributes
        """
        cls.counts[agent.race] -= 1

    def set_stats(self, stats: Dict[str, int], time: int):
        if self.active:
            stats["prep"] += 1

            if self.time == time:
                stats["prep_new"] += 1

            if self.type == "Inj":
                stats["prep_injectable"] += 1
            elif self.type == "Oral":
                stats["prep_oral"] += 1

    def get_acquisition_risk_multiplier(self, time: int, interaction_type: str):
        """
        Get a multiplier for how prep reduces risk of HIV acquisition.

        By default, returns 1.0

        args:
            time: the current model time step
            interaction_type: The type of interaction where the agent could acquire HIV (e.g. 'sex', 'injection' - from [params.classes.interaction_types])
        """
        if self.active and self.last_dose_time is not None:
            params = self.agent.location.params
            if self.type == "Oral":
                adherence = "adherent" if self.adherent else "non_adherent"
                return 1.0 - params.prep.efficacy[adherence]
            elif self.type == "Inj":
                annualized_last_dose_time = (
                    time - self.last_dose_time
                ) / params.model.time.steps_per_year
                annualized_half_life = params.prep.half_life / 365
                load = params.prep.peak_load * (
                    (0.5) ** (annualized_last_dose_time / annualized_half_life)
                )
                return np.exp(-5.528636721 * load)

        return 1.0

    # =============== HELPER METHODS ===================

    def initiate(self, model: "model.TITAN", force: bool = False):
        """
        Place agents onto PrEP treatment. PrEP treatment assumes that the agent knows their HIV status is negative.

        args:
            model : instance of TITAN being run
            force : whether to force the agent to enroll instead of using the appropriate algorithm per the prep params
        """
        # Prep only valid for agents not on prep and are HIV negative
        if self.active or self.agent.hiv.active:  # type: ignore[attr-defined]
            return

        params = self.agent.location.params

        if force:
            self.enroll(model.run_random, model.time)
        elif params.prep.cap_as_prob:
            if "Racial" in params.prep.target_model:
                if (
                    model.run_random.random()
                    <= params.demographics[self.agent.race]
                    .sex_type[self.agent.sex_type]
                    .prep.cap
                ):
                    self.enroll(model.run_random, model.time)
            else:
                if model.run_random.random() <= params.prep.cap:
                    self.enroll(model.run_random, model.time)
        else:
            if "Racial" in params.prep.target_model:
                num_prep_agents = self.counts[self.agent.race]
                all_hiv_agents = exposures.HIV.agents
                all_race = {
                    a for a in model.pop.all_agents if a.race == self.agent.race
                }

                num_hiv_agents = len(all_hiv_agents & all_race)
                target_prep = (len(all_race) - num_hiv_agents) * params.demographics[
                    self.agent.race
                ].sex_type[self.agent.sex_type].prep.cap
            else:
                num_prep_agents = sum(self.counts.values())
                target_prep = int(
                    (model.pop.all_agents.num_members() - len(exposures.HIV.agents))
                    * params.prep.cap
                )

            if num_prep_agents < target_prep:
                self.enroll(model.run_random, model.time)

    def enroll(self, rand_gen, time):
        """
        Enroll an agent in PrEP

        args:
            rand_gen: random number generator
        """
        params = self.agent.location.params

        self.active = True
        self.time = time
        self.last_dose_time = time

        self.adherent = (
            rand_gen.random()
            < params.demographics[self.agent.race]
            .sex_type[self.agent.sex_type]
            .prep.adherence
        )

        if "Inj" in params.prep.type and "Oral" in params.prep.type:
            if rand_gen.random() < params.prep.lai.prob:
                self.type = "Inj"
            else:
                self.type = "Oral"

        else:
            self.type = params.prep.type[0]

        self.add_agent(self.agent)

    def progress(self, model: "model.TITAN", force: bool = False):
        """
        Update agent's PrEP status and discontinue stochastically or if `force` is True

        args:
            model: instance of the TITAN being run
            force: whether to force discontinuation of PrEP
        """
        if force:
            self.discontinue()  # TO_REVIEW should this just remove the agent from counts, or discontinue? does it depend on type?
            return

        if self.type == "Oral":
            if (
                model.run_random.random()
                < self.agent.location.params.demographics[self.agent.race]
                .sex_type[self.agent.sex_type]
                .prep.discontinue
            ):
                self.discontinue()
            else:
                self.last_dose_time = model.time

        # TO_REVIEW should inj prep have a way to continue at the year mark (besides maybe getting prep again through the normal channels of enrollment)?
        if (
            self.type == "Inj"
            and self.last_dose_time
            + self.agent.location.params.model.time.steps_per_year
            == model.time
        ):
            self.discontinue()

    def discontinue(self):
        """
        Discontinue PrEP usage
        """
        self.active = False
        self.type = ""
        self.time = None
        self.last_dose_time = None

        self.remove_agent(self.agent)

    def eligible(self, time) -> bool:
        """
        Determine if an agent is eligible for PrEP

        returns:
            whether the agent is eligible
        """
        params = self.agent.location.params
        if self.agent.hiv.active or time < params.prep.start_time:  # type: ignore[attr-defined]
            return False

        target_model = params.prep.target_model
        gender = params.classes.sex_types[self.agent.sex_type].gender

        if (
            self.active
            or self.agent.vaccine.active  # type: ignore[attr-defined]
            or params.features.random_trial
        ):
            return False

        all_eligible_models = {"Allcomers", "Racial"}

        if all_eligible_models.intersection(target_model):
            return True

        if "cdc_women" in target_model:
            if gender == "F":
                if self.cdc_eligible():
                    return True

        if "cdc_msm" in target_model:
            if gender == "M" and self.cdc_eligible():
                return True

        if "pwid_sex" in target_model:
            if self.agent.drug_type == "Inj" and self.cdc_eligible():
                return True

        if "pwid" in target_model:
            if self.agent.drug_type == "Inj":
                return True

        if "ssp_sex" in target_model:
            if self.agent.syringe_services.active and self.cdc_eligible():  # type: ignore[attr-defined]
                return True

        if "ssp" in target_model:
            if self.agent.syringe_services.active:  # type: ignore[attr-defined]
                return True

        return False

    def cdc_eligible(self) -> bool:
        """
        Determine agent eligibility for PrEP under CDC criteria

        returns:
            cdc eligibility
        """
        if self.agent.is_msm():
            return True

        ongoing_duration = self.agent.location.params.partnership.ongoing_duration
        for rel in self.agent.relationships:
            partner = rel.get_partner(self.agent)
            if rel.duration > ongoing_duration and partner.hiv.dx:  # type: ignore[attr-defined]
                return True

            if partner.drug_type == "Inj" or partner.is_msm():
                return True

        return False



from typing import Dict, ClassVar

from . import base_feature
from .. import agent
from .. import population
from .. import model
from ..parse_params import ObjMap


class HAART(base_feature.BaseFeature):
    """
    Highly Active Antiretroviral Theray (HAART) is a treatment regimen.
    """

    name = "haart"
    stats = ["haart"]
    """
        HAART collects the following stats:

        * haart - number of agents with active haart
    """

    counts: ClassVar[Dict] = {}

    def __init__(self, agent: "agent.Agent"):
        super().__init__(agent)

        self.active = False
        self.ever = False
        self.adherent = False

    @classmethod
    def init_class(cls, params: "ObjMap"):
        """
        Initialize the counts dictionary for the races and sex_types in the model.

        args:
            params: the population params
        """
        cls.counts = {
            race: {sex_type: 0 for sex_type in params.classes.sex_types}
            for race in params.classes.races
        }

    def init_agent(self, pop: "population.Population", time: int):
        """
        Initialize the agent for this feature during population initialization (`Population.create_agent`).  Called on only features that are enabled per the params.

        An Agent can only be initialized with HAART if they are HIV+ and diagnosed.  They are randomly assigned to HAART with a probability based on demographics and then if assigned to haart, assigned an adherence based on those same demographic params.

        args:
            pop: the population this agent is a part of
            time: the current time step
        """
        haart_params = (
            self.agent.location.params.demographics[self.agent.race]
            .sex_type[self.agent.sex_type]
            .drug_type[self.agent.drug_type]
            .haart
        )
        if (
            self.agent.hiv.dx  # type: ignore[attr-defined]
            and pop.pop_random.random() < haart_params.init
        ):
            self.initiate(pop.pop_random, haart_params, "init")

    def update_agent(self, model: "model.TITAN"):
        """
        Update the agent for this feature for a time step.  Called once per time step in `TITAN.update_all_agents`. Agent level updates are done after population level updates.   Called on only features that are enabled per the params.

        Account for HIV treatment through highly active antiretroviral therapy (HAART). HAART was implemented in 1996, hence, there is treatment only after 1996. HIV treatment assumes that the agent knows their HIV+ status (`dx` is True).

        args:
            model: the instance of TITAN currently being run
        """
        if (
            self.agent.hiv.dx  # type: ignore[attr-defined]
            and model.time >= model.params.hiv.start_time  # haart starts with hiv
        ):
            # Determine probability of HIV treatment
            haart_params = (
                self.agent.location.params.demographics[self.agent.race]
                .sex_type[self.agent.sex_type]
                .drug_type[self.agent.drug_type]
                .haart
            )
            # Go on HAART
            if not self.active:
                self.enroll(model, haart_params)

            # Update agents on HAART
            else:
                # Go off HAART
                if model.run_random.random() < haart_params.discontinue:
                    self.active = False
                    self.adherent = False
                    self.remove_agent(self.agent)
                # Become non-adherent
                elif (
                    self.adherent
                    and model.run_random.random() < haart_params.adherence.discontinue
                ):
                    self.adherent = False
                # Become adherent
                elif (
                    not self.adherent
                    and model.run_random.random() < haart_params.adherence.become
                ):
                    self.adherent = True

    @classmethod
    def add_agent(cls, agent: "agent.Agent"):
        """
        Add an agent to the class (not instance).

        Increments `counts` or haart agents by race and sex_type for the given agent.

        args:
            agent: the agent to add to the class attributes
        """
        cls.counts[agent.race][agent.sex_type] += 1

    @classmethod
    def remove_agent(cls, agent: "agent.Agent"):
        """
        Remove an agent from the class (not instance).

        Decrements `counts` or haart agents by race and sex_type for the given agent.

        args:
            agent: the agent to remove from the class attributes
        """
        cls.counts[agent.race][agent.sex_type] -= 1

    def set_stats(self, stats: Dict[str, int], time: int):
        if self.active:
            stats["haart"] += 1

    def get_transmission_risk_multiplier(self, time: int, interaction_type: str):
        """
        Get a multiplier for how haart reduces hiv transmission risk based on interaction type and params.

        By default, returns 1.0

        args:
            time: the current model time step
            interaction_type: The type of interaction where the agent could transmit HIV (e.g. 'sex', 'injection' - from [params.classes.interaction_types])
        """
        prob = 1.0
        if self.active:
            params = self.agent.location.params
            adherence = "adherent" if self.adherent else "non_adherent"
            if interaction_type == "injection":
                prob = params.partnership.injection.transmission.haart_scaling[
                    adherence
                ]
            elif interaction_type == "sex":
                prob = params.partnership.sex.haart_scaling[self.agent.sex_type][
                    adherence
                ]

            # Tuning parameter for ART efficiency
            prob *= params.calibration.haart.transmission

        return prob

    def aids_scale(self):
        prob = 1.0
        if self.active:
            adherence = "adherent" if self.adherent else "non_adherent"
            prob = self.agent.location.params.haart.aids_scale[adherence]

        return prob

    # =========== HELPER METHODS ============

    def enroll(self, model: "model.TITAN", haart_params: ObjMap):
        """
        Determine whether to enroll an agent in HAART.

        args:
            model: the instance of TITAN currently being run
            haart_params: the HAART demographic params for this agent
        """
        if self.agent.location.params.haart.use_cap:
            self.enroll_cap(model, haart_params)
        else:
            self.enroll_prob(model, haart_params)

    def enroll_cap(self, model: "model.TITAN", haart_params: ObjMap):
        """
        Determine whether to enroll an agent in HAART using the cap method.

        args:
            model: the instance of TITAN currently being run
            haart_params: the HAART demographic params for this agent
        """
        race = self.agent.race
        sex_type = self.agent.sex_type
        # if HAART is based on cap instead of prob, determine number of
        # HAART agents based on % of diagnosed agents
        num_dx_agents = self.agent.hiv.dx_counts[race][sex_type]  # type: ignore[attr-defined]
        num_haart_agents = self.counts[race][sex_type]

        # take value from dictionary for cap
        if num_haart_agents < (haart_params.cap * num_dx_agents):
            self.initiate(model.run_random, haart_params, "prob")

    def enroll_prob(self, model: "model.TITAN", haart_params: ObjMap):
        """
        Determine whether to enroll an agent in HAART using probability method.

        args:
            model: the instance of TITAN currently being run
            haart_params: the HAART demographic params for this agent
        """
        if self.ever and self.agent.location.params.haart.use_reinit:
            if model.run_random.random() < haart_params.reinit.prob:
                self.initiate(model.run_random, haart_params, "prob")
        else:
            # Find enroll probability based on time since diagnosis
            enroll_prob = 0.0
            dx_duration = model.time - self.agent.hiv.dx_time  # type: ignore[attr-defined]
            for i in haart_params.enroll.values():
                if i.start <= dx_duration < i.stop:
                    enroll_prob = i.prob * model.calibration.haart.coverage
                    break

            if model.run_random.random() < (enroll_prob):
                self.initiate(model.run_random, haart_params, "prob")

    def initiate(self, rand_gen, haart_params: ObjMap, init_or_prob: str):
        """
        Initiate an agent with HAART and add them to the population.

        args:
            model: the instance of TITAN currently being run
        """
        self.adherent = rand_gen.random() < haart_params.adherence[init_or_prob]

        # Add agent to HAART class set, update agent params
        self.active = True
        self.ever = True
        self.add_agent(self.agent)



import paraml  # type: ignore

import os
from inspect import getsourcefile
from pathlib import Path
import math
from typing import Optional, Dict
from copy import deepcopy


class ObjMap(dict):
    """
    A dictionary-like class which allows accessing members either using standard
    dictionary notation or dots.  Note the hash function is hard-coded - beware.
    """

    def __init__(self, d: Dict):
        for k, v in d.items():
            if isinstance(v, dict):
                v = self.__class__(v)
            self[k] = v

    def __getattribute__(self, k):
        try:
            return self[k]
        except KeyError:
            return object.__getattribute__(self, k)

    def __setattr__(self, k, v):
        return self.__setitem__(k, v)

    def __hash__(self):
        return 1234567890

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.items():
            result[k] = deepcopy(v, memo)
        return result


# ============== PARSING FUNCTIONS ======================


def check_params(params: ObjMap):
    """
    Consistency checks for param populations
    """
    race_pop = 0
    for race in params.classes.races:
        r_dems = params.demographics[race]
        race_pop += r_dems.ppl
        sex_type_pop = 0
        for st, st_dems in r_dems.sex_type.items():
            if st in list(params.classes.sex_types.keys()):
                sex_type_pop += st_dems.ppl

            drug_type_pop = 0
            for dt, dt_dems in st_dems.drug_type.items():
                if dt in list(params.classes.drug_types):
                    drug_type_pop += dt_dems.ppl

            assert math.isclose(
                drug_type_pop, 1, abs_tol=0.001
            ), f"ppl of {race}'s {st}'s drug_types must add to 1. Currently adding to {drug_type_pop}"

        assert math.isclose(
            sex_type_pop, 1, abs_tol=0.001
        ), f"ppl of {race}'s sex_types must add to 1. Currently adding to {sex_type_pop}"

    assert math.isclose(race_pop, 1, abs_tol=0.001), "ppl of races must add to 1"

    loc_pop = 0
    for location in params.classes.locations.values():
        loc_pop += location.ppl

    assert math.isclose(loc_pop, 1, abs_tol=0.001), "ppl of locations must add to 1"

    for param, assort in params.assort_mix.items():
        assort_value = 0
        for ptnr_value in assort.partner_values.values():
            assort_value += ptnr_value
        assert math.isclose(
            assort_value, 1, abs_tol=0.001
        ), f"assort values must add to 1, not {assort_value} in {param}"


def create_params(
    setting_name: Optional[str],
    param_path: str,
    outdir: str,
    error_on_unused: bool = False,
) -> ObjMap:
    """
    Entry function - given the path to the setting, params, output directory and whether
    or not to use the base setting. Parse and create a params (ObjMap) object.

    args:
        setting_name: path to a settings file or directory or `None`
        param_path: path to parameter file or directory
        outdir: path to directory where computed params will be saved
        error_on_unused: throw a hard error if there are unused parameters, otherwise warnings are only printed

    returns:
        computed/validated model paramters with defaults filled in where needed
    """
    # find defs, where we are in the code for settings and base
    filename = getsourcefile(create_params)  # what is the sourcefile for this function
    if filename is not None:
        parent = Path(filename).resolve().parent
    else:
        raise Exception("can't find where I am in the code?")

    param_defs = os.path.join(parent, "params")

    param_paths = []

    # merge setting and params
    if setting_name is not None:
        # check if it's a known setting or pass it through as a path
        if setting_name in os.listdir(os.path.join(parent, "settings")):
            param_paths.append(os.path.join(parent, "settings", setting_name))
        else:
            param_paths.append(setting_name)

    param_paths.append(param_path)

    parsed = paraml.create_params(
        param_defs,
        *param_paths,
        out_path=os.path.join(outdir, "params.yml"),
        error_on_unused=error_on_unused,
    )

    parsed = ObjMap(parsed)
    check_params(parsed)

    return parsed



from . import base_interaction
from .. import model
from .. import agent
from .. import utils


class PCA(base_interaction.BaseInteraction):

    name = "pca"

    @classmethod
    def get_num_acts(cls, model: "model.TITAN", rel: "agent.Relationship") -> int:
        params = model.params.partnership.pca.frequency[rel.bond_type]

        if params.type == "bins":
            params = params.bins
            acts_prob = model.run_random.random()
            acts_bin = 1
            current_p_value = params[acts_bin].prob

            while acts_prob > current_p_value:
                acts_bin += 1
                current_p_value += params[acts_bin].prob

            min = params[acts_bin].min
            max = params[acts_bin].max

            if min == max:
                num_acts = min
            else:
                num_acts = model.run_random.randrange(min, max)
        elif params.type == "distribution":
            num_acts = round(utils.safe_dist(params.distribution, model.run_random))

        return num_acts



from . import base_interaction
from .. import utils
from .. import model
from .. import agent


class Sex(base_interaction.BaseInteraction):

    name = "sex"

    @classmethod
    def get_num_acts(cls, model: "model.TITAN", rel: "agent.Relationship") -> int:
        """
        Simulate random transmission of HIV between two agents through Sex. One of the agents must be HIV+.

        args:
            model: The model being run
            rel : Relationship
        """
        # unprotected sex probabilities for primary partnerships
        mean_sex_acts = (
            rel.get_number_of_sex_acts(model.np_random) * model.calibration.sex.act
        )
        total_sex_acts = utils.poisson(mean_sex_acts, model.np_random)

        # Get condom usage
        p_safe_sex = (
            rel.agent1.location.params.demographics[rel.agent1.race]
            .sex_type[rel.agent1.sex_type]
            .safe_sex[rel.bond_type]
            .prob
        )

        # increase condom usage if diagnosed
        if rel.agent1.hiv.dx or rel.agent2.hiv.dx:  # type: ignore[attr-defined]
            # Calculate probability of safe sex given risk reduction
            p_unsafe_sex = (1 - p_safe_sex) * (
                1 - model.params.hiv.dx.risk_reduction.sex
            )
            p_safe_sex = 1 - p_unsafe_sex

        # Reduction of risk acts between partners for condom usage
        unsafe_sex_acts = total_sex_acts
        for n in range(unsafe_sex_acts):
            if model.run_random.random() < p_safe_sex:
                unsafe_sex_acts -= 1

        return unsafe_sex_acts



# mypy: always-true=HighRisk

from typing import Dict, Optional

from . import base_feature
from .. import utils
from .. import agent
from .. import population
from .. import model


class HighRisk(base_feature.BaseFeature):

    name = "high_risk"
    stats = [
        "high_risk_new",
        "high_risk_new_hiv",
        "high_risk_new_aids",
        "high_risk_new_dx",
        "high_risk_new_haart",
        "hiv_new_high_risk",
        "hiv_new_high_risk_ever",
    ]
    """
        High Risk collects the following stats:

        * high_risk_new - number of agents that became active high risk this time step
        * high_risk_new_hiv - number of agents that became active high risk this time step with HIV
        * high_risk_new_aids - number of agents that became active high risk this time step with AIDS
        * high_risk_new_dx - number of agents that became active high risk this time step with diagnosed HIV
        * high_risk_new_haart - number of agents that became active high risk this time step with active HAART
        * hiv_new_high_risk - number of agents that became active with HIV this time step who are high risk
        * hiv_new_high_risk_ever - number of agents that became active with HIV this time step were ever high risk
    """

    def __init__(self, agent: "agent.Agent"):
        super().__init__(agent)

        self.active = False
        self.time: Optional[int] = None
        self.duration = 0
        self.ever = False

    def init_agent(self, pop: "population.Population", time: int):
        """
        Initialize the agent for this feature during population initialization (`Population.create_agent`).  Called on only features that are enabled per the params.

        Based on agent demographic params, randomly initialize agent as high risk.

        args:
            pop: the population this agent is a part of
            time: the current time step
        """
        if (
            pop.pop_random.random()
            < self.agent.location.params.demographics[self.agent.race]
            .sex_type[self.agent.sex_type]
            .high_risk.init
        ):
            self.become_high_risk(pop, time)

    def update_agent(self, model: "model.TITAN"):
        """
        Update the agent for this feature for a time step.  Called once per time step in `TITAN.update_all_agents`. Agent level updates are done after population level updates.   Called on only features that are enabled per the params.

        Update high risk agents or remove them from high risk pool.  An agent becomes high_risk through the incarceration feature

        args:
            model: the instance of TITAN currently being run
        """
        if not self.active:
            # released last step, evaluate agent for high risk
            if self.agent.incar.release_time == model.time - 1:  # type: ignore[attr-defined]
                self.become_high_risk(model.pop, model.time)

            # incarcerated last step, evaluate agent's partners for high risk
            elif self.agent.incar.time == model.time - 1:  # type: ignore[attr-defined]

                # put partners in high risk
                for partner in self.agent.get_partners(
                    self.agent.location.params.high_risk.partnership_types
                ):
                    if (
                        not partner.high_risk.active  # type: ignore[attr-defined]
                        and model.run_random.random()
                        < partner.location.params.high_risk.prob
                    ):
                        partner.high_risk.become_high_risk(model.pop, model.time)  # type: ignore[attr-defined]
        elif self.duration > 0:
            self.duration -= 1
        else:
            self.active = False

            self.update_partner_numbers(
                model.pop, -1 * self.agent.location.params.high_risk.partner_scale
            )

            for bond in self.agent.location.params.high_risk.partnership_types:
                num_ended = 0
                while (
                    len(self.agent.partners[bond]) - num_ended
                ) > self.agent.target_partners[bond]:
                    rel = utils.safe_random_choice(
                        self.agent.relationships, model.run_random
                    )
                    if rel is not None:
                        num_ended += 1
                        rel.duration = 0  # will end on next step

    def set_stats(self, stats: Dict[str, int], time: int):
        if self.time == time:
            stats["high_risk_new"] += 1
            if self.agent.hiv.active:  # type: ignore[attr-defined]
                stats["high_risk_new_hiv"] += 1
                if self.agent.hiv.aids:  # type: ignore[attr-defined]
                    stats["high_risk_new_aids"] += 1
                if self.agent.hiv.dx:  # type: ignore[attr-defined]
                    stats["high_risk_new_dx"] += 1
                    if self.agent.haart.active:  # type: ignore[attr-defined]
                        stats["high_risk_new_haart"] += 1

        # newly hiv
        if self.agent.hiv.time == time:  # type: ignore[attr-defined]
            if self.active:
                stats["hiv_new_high_risk"] += 1
            if self.ever:
                stats["hiv_new_high_risk_ever"] += 1

    # ============== HELPER METHODS ================

    def become_high_risk(
        self, pop: "population.Population", time: int, duration: int = None
    ):
        """
        Mark an agent as high risk and assign a duration to their high risk period

        args:
            pop: the model poopulation
            time: the time step the agent is becoming high risk
            duration: duration of the high risk period, defaults to param value if not passed [params.high_risk.sex_based]
        """

        if not self.agent.location.params.features.high_risk:
            return None

        if not self.ever:
            self.time = time

        self.active = True
        self.ever = True

        if duration is not None:
            self.duration = duration
        else:
            self.duration = self.agent.location.params.high_risk.sex_based[
                self.agent.sex_type
            ].duration

        self.update_partner_numbers(
            pop, self.agent.location.params.high_risk.partner_scale
        )

    def update_partner_numbers(self, pop: "population.Population", amount: int):
        """
        Update the agent's mean and target partner numbers by the amount passed.  Update partnerability for the population.

        args:
            pop: the model population
            amount: the positive or negatative amount to adjust the mean by
        """
        for bond in self.agent.location.params.high_risk.partnership_types:
            self.agent.mean_num_partners[bond] += amount  # could be negative
            self.agent.target_partners[bond] = utils.poisson(
                self.agent.mean_num_partners[bond], pop.np_random
            )
            pop.update_partnerability(self.agent)



from typing import Optional, Set, Dict, List, Any
from copy import deepcopy
import math

from .parse_params import ObjMap
from . import utils


class Location:
    def __init__(self, name: str, defn: ObjMap, params: ObjMap):
        """
        This class constructs and represents a location within the model.  A location
            can have an arbitrary geographic granularity.

        args:
            name: name of the location
            defn: definition for this location
            params: model parameters
        """
        # location properties
        self.name = name
        self.params = self.create_params(params)
        self.ppl = defn.ppl  # percent of overall population assigned to this location

        # value/weight maps needed for creating new agents in this location
        self.pop_weights: Dict[str, Dict[str, List[Any]]] = {}
        self.role_weights: Dict[str, Dict] = {}
        self.drug_weights: Dict[str, Dict] = {}
        self.init_weights()

        self.edges: Set["LocationEdge"] = set({})  # or maybe edges instead

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"'{self.name}'"

    def __eq__(self, other):
        return self.name == other.name

    def __ne__(self, other):
        return self.name != other.name

    def __hash__(self):
        return hash(self.name)

    def create_params(self, params: ObjMap) -> ObjMap:
        """
        Scale or override the generic parameters with any location based scaling from params.location.scaling

        args:
            params: model parameters

        returns:
            new parameter object with scaled values for this location
        """
        new_params = deepcopy(params)

        defns = new_params.location.scaling[self.name]
        for param_path, defn in defns.items():
            if param_path != "ls_default":
                if defn.field == "scalar":
                    utils.scale_param(new_params, param_path, defn.scalar)
                elif defn.field == "override":
                    utils.override_param(new_params, param_path, defn.override)

        return new_params

    def init_weights(self):
        """
        Create the containers to hold values and weights for randomly selecting:

        * sex_role
        * drug_type
        * race
        * sex_type
        """

        def init_weight_dict(d, item):
            d[item] = {}
            d[item]["values"] = []
            d[item]["weights"] = []

        def add_weight(d, v, w):
            d["values"].append(v)
            d["weights"].append(w)

        total_ppl = 0
        for race, race_param in self.params.demographics.items():
            self.role_weights[race] = {}
            self.drug_weights[race] = {}
            init_weight_dict(self.pop_weights, race)
            total_ppl += race_param.ppl
            for st, st_param in race_param.sex_type.items():
                add_weight(self.pop_weights[race], st, st_param.ppl)
                init_weight_dict(self.role_weights[race], st)
                init_weight_dict(self.drug_weights[race], st)
                for role, prob in st_param.sex_role.init.items():
                    add_weight(self.role_weights[race][st], role, prob)
                for dt, dt_param in st_param.drug_type.items():
                    add_weight(self.drug_weights[race][st], dt, dt_param.ppl)

                assert math.isclose(
                    sum(self.role_weights[race][st]["weights"]), 1, abs_tol=0.001
                ), f"{self.name}'s' {race} {st} role weights must add to 1"
                assert math.isclose(
                    sum(self.drug_weights[race][st]["weights"]), 1, abs_tol=0.001
                ), f"ppl of {self.name}'s' {race} {st} drug_types must add to 1"

            assert math.isclose(
                sum(self.pop_weights[race]["weights"]), 1, abs_tol=0.001
            ), f"ppl of {self.name}'s' {race} sex_types must add to 1"

        assert math.isclose(
            total_ppl, 1, abs_tol=0.001
        ), f"ppl of {self.name}'s' races must add to 1"


# LocationEdges are very much a WIP and not actually used anywhere yet
# outstanding questions:
# * should edges be directed? bi-drectional? uni-drectional, but both directions housed in the same edge?
# * what attributes do edges need?
# * how will mobility be implemented?
# * assorting?
class LocationEdge:

    next_edge_id = 0

    @classmethod
    def update_id_counter(cls, last_id: int):
        cls.next_edge_id = last_id + 1

    def __init__(
        self, loc1: Location, loc2: Location, distance: float, id: Optional[int] = None
    ):
        """
        Construct a location edge, which holds attributes that relate two Locations.

        args:
            loc1: the first location
            loc2: the other location
            distance: a measure of distance between the locations
            id: a unique identifier for this edge
        """
        assert loc1 != loc2, "can't have a location self-edge"

        # self.id is unique ID number used to track each edge.
        if id is not None:
            self.id = id
        else:
            self.id = self.next_edge_id

        self.update_id_counter(self.id)

        self.edge = set({loc1, loc2})
        self.distance = distance


class Geography:
    def __init__(self, params: ObjMap):
        """
        Umbrella class to initialize/store locations and location edges for a population

        args:
            params: model parameters
        """

        self.locations: Dict[str, Location] = {
            location: Location(location, defn, params)
            for location, defn in params.classes.locations.items()
        }

        self.edges: Set[LocationEdge] = set()
        for name, defn in params.location.edges.items():
            if name != "edge_default":
                loc1 = self.locations[defn.location_1]
                loc2 = self.locations[defn.location_2]
                self.edges.add(LocationEdge(loc1, loc2, defn.distance))



from typing import List, Dict

from .. import agent
from .. import population
from .. import model


class BaseExposure:

    name: str = ""
    """Name of exposure in the params file.  Also used to name the attribute in Agent"""

    stats: List[str] = []
    """List of names of stats that come from this exposure (e.g. hiv.dx)"""

    def __init__(self, agent: "agent.Agent"):
        self.active = False
        self.agent = agent

    @classmethod
    def init_class(cls, params):
        """
        Initialize any class level attributes (such as setting counters to zero). Called on every active exposure on population initialization.

        args:
            params: parameters for this population
        """
        pass

    def init_agent(self, pop: "population.Population", time: int):
        """
        Initialize the agent for this exposure during population initialization (`Population.create_agent`).  Called only on exposures that are enabled per the params.

        args:
            pop: the population this agent is a part of
            time: the current time step
        """
        pass

    def update_agent(self, model: "model.TITAN"):
        """
        Update the agent for this exposure for a time step.  Called once per time step in `TITAN.update_all_agents`.

        args:
            model: the instance of TITAN currently being run
        """
        pass

    @classmethod
    def add_agent(cls, agent: "agent.Agent"):
        """
        Add an agent to the class (not instance).  This can be useful if tracking population level statistics or groups, such as counts or newly active agents.

        This method is not called from anywhere in the model, but creates a cohesive api with `remove_agent`, which is called from `Population.remove_agent`.

        args:
            agent: the agent to add to the class attributes
        """
        pass

    @classmethod
    def remove_agent(cls, agent: "agent.Agent"):
        """
        Remove an agent from the class (not instance).  This can be useful if tracking population level statistics or groups, such as counts.

        This method is called from `Population.remove_agent`, but may also need to be called within the feature if an agent transitions from `active == True` to `active == False`.

        args:
            agent: the agent to remove from the class attributes
        """
        pass

    def set_stats(self, stats: Dict[str, int], time: int):
        """
        Update the `stats` dictionary passed for this agent.  Called from `output.get_stats` for each enabled exposure in the model.

        The stats to be updated must be declared in the class attribute `stats` to make sure the dictionary has the expected keys/counter value initialized.

        args:
            stats: the dictionary to update with this agent's feature statistics
            time: the time step of the model when the stats are set
        """
        pass

    @staticmethod
    def expose(
        model: "model.TITAN",
        interaction: str,
        rel: "agent.Relationship",
        num_acts: int,
    ):
        """
        Expose a relationship to the exposure for a number of acts for a specific interaction type.  Typically, this determines if the exposure can cause conversion/change in one of the agents, then if so, determines the probability of that and then converts the succeptible agent.

        args:
            model: The running model
            interaction: The type of interaction (e.g. sex, injection)
            rel: The relationship where the interaction is occuring
            num_acts: The number of acts of that interaction
        """
        pass

    def get_transmission_probability(
        self,
        model: "model.TITAN",
        interaction: str,
        partner: "agent.Agent",
        num_acts: int,
    ) -> float:
        """
        Determines the probability of a transmission event from agent to partner based on
            interaction type.

        This is not called from anywhere else in the model, but is recommended as a way to structure the exposure method into getting the transmission probability, then doing the conversion.

        args:
            model: The running model
            interaction : The type of interaction (e.g. `sex`, `injection`)
            partner: The agent's partner
            num_acts: The number of acts where exposure occured

        returns:
            probability of transmission from agent to partner
        """
        return 0.0

    def convert(self, model: "model.TITAN"):
        """
        Convert the agent to the exposure (i.e. become active).

        args:
            model: The running model
        """
        pass

    def diagnose(self, model: "model.TITAN"):
        """
        Diagnose the agent with the exposure (if applicable).

        args:
            model: The running model
        """
        pass



import random
from functools import wraps
from typing import TypeVar, Collection, Union, Iterable
from math import floor
import logging
import os
from datetime import datetime

import networkx as nx  # type: ignore

from . import distributions
from .parse_params import ObjMap


def memo(f):
    """
    Decorator to memoize a function
    (caches results given args, only use if deterministic)
    """
    cache = {}

    @wraps(f)
    def wrap(*arg):
        if arg not in cache:
            cache[arg] = f(*arg)
        return cache[arg]

    return wrap


def get_check_rand_int(seed: int) -> int:
    """
    Check the value passed of a seed, make sure it's an int, if 0, get a random seed

    args:
        seed: integer to check or replace with a seed

    returns:
        validated seed
    """
    if type(seed) is not int or seed < 0:
        raise ValueError("Random seed must be positive integer")
    elif seed == 0:
        return random.randint(1, 1000000)
    else:
        return seed


def safe_divide(numerator: int, denominator: int) -> float:
    """
    Divide two numbers, but default 0 if denominator is 0, otherwise divide as normal.

    args:
        numerator: number being divided
        denominator: number doing the dividing

    returns:
        resulting number
    """
    if denominator == 0:
        return 0.0
    else:
        return 1.0 * numerator / denominator


# Requirement for safe_random_choice function
T = TypeVar("T")


def safe_random_choice(seq, rand_gen, weights=None):
    """
    Return None or a random choice from a collection of items

    args:
        seq: collection to select a random item from
        rand_gen: random number generator
        weights: an optional collection of weights to use instead of a uniform distribution

    returns:
        an item, or `None` if the collection is empty
    """
    if not seq:
        return None

    if isinstance(seq, set):
        seq = tuple(seq)

    # don't call out to random choices if we don't need to (for performance)
    if len(seq) == 1:
        return seq[0]
    elif len(seq) == 2 and weights is None:
        return seq[0] if rand_gen.random() <= 0.5 else seq[1]

    choices = rand_gen.choices(seq, weights=weights)
    return choices[0]


def safe_random_int(start: int, stop: int, rand_gen) -> int:
    """
    Return an integer between [start, stop)

    args:
        start: start value
        stop: stop value
        rand_gen: random number generator

    returns:
        an item, or `None` if the collection is empty
    """
    return floor(rand_gen.random() * (stop - start) + start)


def safe_shuffle(seq: Collection[T], rand_gen) -> Iterable[T]:
    """
    Return None or a shuffled sequence

    args:
        seq: collection to shuffle
        rand_gen: random number generator

    returns:
        shuffled sequence, or `None` if empty
    """
    if seq:
        if isinstance(seq, set):
            seq = list(seq)
        rand_gen.shuffle(seq)
        return seq
    else:
        return []


@memo
def parse_var(dist_value, dist_type):
    type_caster = eval(dist_type)
    return type_caster(dist_value)


@memo
def get_dist(rand_gen, dist_type):
    if dist_type == "randint":
        return lambda *args: safe_random_int(*args, rand_gen)
    elif hasattr(rand_gen, dist_type):
        return getattr(rand_gen, dist_type)
    elif hasattr(distributions, dist_type):
        dist = getattr(distributions, dist_type)
        return lambda *args: dist(rand_gen, *args)
    else:
        raise AttributeError(f"Distribution type {dist_type} not found!")


def safe_dist(dist_info: ObjMap, rand_gen) -> Union[int, float]:
    """
    Draw a value from a distribution as defined in `dist_info`.

    args:
        dist_info: a definition of a distribution to use [params.classes.distributions]
        rand_gen: random number generator

    returns:
        a value drawn from the distribution
    """
    # gather arguments
    args = []
    for d in dist_info.vars.values():
        args.append(parse_var(d.value, d.value_type))

    dist = get_dist(rand_gen, dist_info.dist_type)
    value = dist(*args)

    if hasattr(value, "__iter__"):  # check if value is any type of sequence
        return value[0]
    else:
        return value


def binom_0(n: int, p: float):
    """
    Mirrors scipy binom.pmf as used in code
    """
    return (1 - p) ** n


def poisson(mu: float, np_rand):
    """
    Mirrors scipy poisson.rvs function as used in code
    """
    return np_rand.poisson(mu)


def get_param_from_path(params: ObjMap, param_path: str, delimiter: str):
    """
    Given a params object and a delimited path, get the leaf of the params tree
    and the last key to access it
    """
    path = param_path.split(delimiter)
    path_params = params
    for p in path[:-1]:
        try:
            path_params = path_params[p]
        except KeyError:
            path_params = path_params[int(p)]

    return path_params, path[-1]


def scale_param(params: ObjMap, param_path: str, scalar: float, delimiter="|"):
    """
    Given the params and a parameter path in the format prep|cap, scale the
    current value by the scalar
    """
    scaling_item, last_key = get_param_from_path(params, param_path, delimiter)

    old_val = scaling_item[last_key]
    logging.info(f"scaling - {param_path}: {old_val} => {old_val * scalar}")
    scaling_item[last_key] = old_val * scalar


def override_param(params: ObjMap, param_path: str, value, delimiter="|"):
    """
    Given the params and a parameter path in the format prep|cap, change the
    current value to new value
    """
    override_item, last_key = get_param_from_path(params, param_path, delimiter)
    try:
        old_val = override_item[last_key]
    except KeyError:
        last_key = int(last_key)
        old_val = override_item[last_key]

    logging.info(f"overriding - {param_path}: {old_val} => {value}")
    override_item[last_key] = value


def total_probability(p: float, num_acts: int) -> float:
    """
    Given a per act probability and a number of acts, return the total probability.

    args:
        p: the per act probability
        num_acts: the number of acts

    returns:
        the total probability
    """
    if num_acts == 1:
        return p
    elif num_acts >= 1:
        return 1.0 - binom_0(num_acts, p)
    else:
        return 0.0


def connected_components(graph):
    """
    Get connected components in graph

    args:
        graph: the model's underlying graph

    returns:
        list of connected components
    """
    return sorted(
        list(graph.subgraph(c) for c in nx.connected_components(graph)),
        key=len,
        reverse=True,
    )


def get_independent_bin(rand_gen, bin_def: ObjMap) -> int:
    """
    Get the bin key given independent bins.  A probability is selected at random, then each bin's `prob` is compared to it, the first bin that has a `prob` less than or equal to that probability is returned.

    args:
        rand_gen: A random number generator
        bin_def: The ObjMap containing the bins

    returns:
        The integer key of the matched bin (or last bin if no matches)
    """
    rand_val = rand_gen.random()
    for bin, fields in bin_def.items():
        if rand_val <= fields.prob:
            break

    return bin


def get_cumulative_bin(rand_gen, bin_def: ObjMap) -> int:
    """
    Get the bin key given cumulative bins.  A probability is selected at random, then each bin's `prob` is compared to it, the first bin that has a cumulative `prob` (e.g. for bin 2, the prob of bin 1 plus the prob of bin 2) less than or equal to that probability is returned.

    args:
        rand_gen: random number generator
        bin_def: ObjMap containing the bins

    returns:
        integer key of the matched bin (or last bin if no matches)
    """
    rand_val = rand_gen.random()
    p = 0.0
    for bin, fields in bin_def.items():
        p += fields.prob
        if rand_val <= p:
            break

    return bin


def set_up_logging(params):
    # set up logging
    if params.outputs.logging.destination == "file":
        log_msg_format = "{message:<64}  [{asctime}][{levelname}][{module}]"
        log_dt_format = "%Y%m%d%H%M%S"
        path = (
            os.getcwd()
            if params.outputs.logging.filepath == "__cwd__"
            else params.outputs.logging.filepath
        )
        logging.basicConfig(
            format=log_msg_format,
            style="{",
            datefmt=log_dt_format,
            filename=os.path.join(
                path, f"titan_log_{datetime.now().strftime(log_dt_format)}.txt"
            ),
            level=params.outputs.logging.level,
        )
    else:
        log_msg_format = "%(message)s"
        logging.basicConfig(
            format=log_msg_format,
            level=params.outputs.logging.level,
        )



#!/usr/bin/env python
# encoding: utf-8

# Imports
from typing import Optional, Dict, Set
from copy import copy

import numpy as np  # type: ignore

from . import agent  # import Agent, AgentSet
from . import utils
from . import parse_params


def select_partner(
    agent: "agent.Agent",
    partnerable_agents: Set["agent.Agent"],
    sex_partners: Dict,
    pwid_agents: "agent.AgentSet",
    params: "parse_params.ObjMap",
    rand_gen,
    bond_type: str,
) -> Optional["agent.Agent"]:
    """
    Get a partner for the agent.

    args:
        agent : agent in need of a partner
        partnerable_agents: agents that can be selected as a partner
        sex_partners: mapping from sex_type to agents in the population that can sleep with that sex_type
        pwid_agents: agents with `drug_type==="Inj"`
        params: model parameters
        rand_gen: random number generator
        bond_type: type of relationship that is being formed with the partner

    returns:
        new partner or `None`
    """
    eligible = copy(partnerable_agents)
    eligible -= agent.get_partners()
    eligible -= {agent}

    acts_allowed = params.classes.bond_types[bond_type].acts_allowed

    if "injection" in acts_allowed:
        eligible &= pwid_agents.members

    if "sex" in acts_allowed:
        eligible &= sex_partners[agent.sex_type]

    # short circuit to avoid attempting to assort with no eligible partners
    if not eligible:
        return None

    if params.features.assort_mix:
        match_fns = get_match_fns(
            params.assort_mix.values(), agent, bond_type, rand_gen
        )
        # if no definitions match this agent, don't try to assort
        if len(match_fns) > 0:
            for partner in utils.safe_shuffle(eligible, rand_gen):
                if is_assortable(partner, match_fns):
                    return partner
            return None

    return utils.safe_random_choice(eligible, rand_gen)


# does an agent match the criteria of the randomly chosen assort values?
def is_assortable(agent, match_fns):
    for match_fn in match_fns:
        if not match_fn(agent):
            return False

    return True


# get the attribute as a string, recurse if nested
def get_str_attr(obj, attr):
    attrs = attr.split(".")  # returns list of all attrs
    if len(attrs) == 1:
        return str(getattr(obj, attrs[0]))
    else:
        return get_str_attr(getattr(obj, attrs.pop(0)), ".".join(attrs))


# if this assort rule applies for this agent, get the match function for a potential partner
def get_match_fns(assort_defs, agent, bond_type, rand_gen):
    match_fns = []
    for assort_def in assort_defs:
        if len(assort_def.bond_types) == 0 or bond_type in assort_def.bond_types:
            if assort_def.agent_value == "__any__":
                match_fns.append(get_same_match_fn(assort_def, agent, rand_gen))
            elif get_str_attr(agent, assort_def.attribute) == str(
                assort_def.agent_value
            ):
                match_fns.append(get_match_fn(assort_def, rand_gen))

    return match_fns


# pick a partner type randomly given the weights
def get_partner_type(assort_def, rand_gen):
    partner_types = list(assort_def.partner_values.keys())
    partner_weights = [assort_def.partner_values[p] for p in partner_types]
    return utils.safe_random_choice(partner_types, rand_gen, weights=partner_weights)


# what partner attribute to use in assorting
def get_partner_attr(assort_def):
    if assort_def.partner_attribute == "__agent__":
        return assort_def.attribute
    else:
        return assort_def.partner_attribute


# given an assort def, randomly select the type the partner must have given the
# weights and return a function to determin if a potential partner matches it
def get_match_fn(assort_def, rand_gen):
    partner_types = list(assort_def.partner_values.keys())
    partner_type = get_partner_type(assort_def, rand_gen)
    attr = get_partner_attr(assort_def)
    if partner_type == "__other__":
        partner_types.remove("__other__")
        return lambda ag: get_str_attr(ag, attr) not in partner_types
    else:
        return lambda ag: get_str_attr(ag, attr) == partner_type


# given an assort def where agent_value == '__any__', return the match function
def get_same_match_fn(assort_def, agent, rand_gen):
    partner_type = get_partner_type(assort_def, rand_gen)
    attr = get_partner_attr(assort_def)

    agent_attribute = get_str_attr(agent, attr)
    if partner_type == "__same__":
        return lambda ag: get_str_attr(ag, attr) == agent_attribute
    elif partner_type == "__other__":
        return lambda ag: get_str_attr(ag, attr) != agent_attribute
    else:
        raise ValueError(
            "When using same-assorting, only valid partner_types are __same__ and __other__"
        )


@utils.memo
def sex_possible(
    agent_sex_type: str, partner_sex_type: str, sex_types: "parse_params.ObjMap"
) -> bool:
    """
    Determine if sex is possible.

    args:
        agent_sex_type: name of agent's sex type
        partner_sex_type: name of partner's sex type
        sex_types: params defining in scope sex types

    returns:
        whether sex is possible between agents of these sex types
    """

    # Check input
    if agent_sex_type not in sex_types:
        raise ValueError(f"Invalid agent_sex_type! {agent_sex_type}")
    if partner_sex_type not in sex_types:
        raise ValueError(f"Invalid partner_sex_type! {partner_sex_type}")

    agent_match = agent_sex_type in sex_types[partner_sex_type].sleeps_with
    partner_match = partner_sex_type in sex_types[agent_sex_type].sleeps_with

    return agent_match and partner_match


def get_mean_rel_duration(params: "parse_params.ObjMap"):
    """
    Find the average partnership duration by bond type

    args:
        params: The current model's parameters
    """
    mean_rel_duration: Dict[str, Dict] = {}
    for bond in params.partnership.duration:
        mean_rel_duration[bond] = {}
        for race in params.classes.races:
            if params.partnership.duration[bond][race].type == "bins":
                weights = []
                vals = []
                dur_bins = params.partnership.duration[bond][race].bins
                for bins in dur_bins:
                    if bins > 1:
                        weights.append(dur_bins[bins].prob - dur_bins[bins - 1].prob)
                    else:
                        weights.append(dur_bins[bins].prob)
                    vals.append(np.average([dur_bins[bins].min, dur_bins[bins].max]))
                mean_rel_duration[bond][race] = np.average(vals, weights=weights)
            else:
                mean_rel_duration[bond][race] = params.partnership.duration[bond][
                    race
                ].distribution.mean
            assert (
                mean_rel_duration[bond][race] > 0
            ), "All bonds must have a positive duration!"

    return mean_rel_duration


def get_partnership_duration(
    params: "parse_params.ObjMap", rand_gen, bond_type: str, race: Optional[str]
) -> int:
    """
    Get duration of a relationship drawn from bins or a distribution per the params [params.partnership.duration]

    args:
        params: model parameters
        rand_gen: np random number generator
        bond_type: type of bond for the relationship whose duration is being determined

    returns:
        number of time steps the partnership should endure
    """

    if params.partnership.duration[bond_type][race].type == "bins":
        dur_info = params.partnership.duration[bond_type][race].bins
        i = utils.get_independent_bin(rand_gen, dur_info)
        duration = utils.safe_random_int(dur_info[i].min, dur_info[i].max, rand_gen)

    else:
        dist = params.partnership.duration[bond_type][race].distribution
        duration = int(utils.safe_dist(dist, rand_gen))

    return duration



from typing import List, Dict, Optional, Set

from . import base_exposure
from .. import agent
from .. import population
from .. import model
from .. import utils


class HIV(base_exposure.BaseExposure):

    name: str = "hiv"
    stats: List[str] = ["hiv", "hiv_dx", "hiv_aids", "hiv_new", "hiv_dx_new"]
    """
        HIV collects the following stats:

        * hiv - number of agents with active hiv
        * hiv_dx - number of agents with diagnosed hiv
        * hiv_aids - number of agents with aids
        * hiv_new - number of agents converted to hiv this timestep
        * hiv_dx_new - number of agents with diagnosed with hiv this timestep
    """

    dx_counts: Dict[str, Dict[str, int]] = {}
    """Counts of diagnosed agents by race and sex_type"""

    agents: Set["agent.Agent"] = set()
    """Agents with active hiv"""

    def __init__(self, agent: "agent.Agent"):
        super().__init__(agent)

        self.active = False
        self.time: Optional[int] = None
        self.dx = False
        self.dx_time: Optional[int] = None
        self.aids = False

    @classmethod
    def init_class(cls, params):
        """
        Initialize any diagnosis counts and the agents set.

        args:
            params: parameters for this population
        """
        cls.dx_counts = {
            race: {so: 0 for so in params.classes.sex_types}
            for race in params.classes.races
        }
        cls.agents = set()

    def init_agent(self, pop: "population.Population", time: int):
        """
        Initialize the agent for this exposure during population initialization (`Population.create_agent`).  Called on only exposures that are enabled per the params.

        Based on demographic params for the agent, stochastically determine if hiv is active, and if active, at what past time point was the agent converted, if the agent is diagnosed, and if the agent has aids.

        args:
            pop: the population this agent is a part of
            time: the current time step
        """
        agent_params = (
            self.agent.location.params.demographics[self.agent.race]
            .sex_type[self.agent.sex_type]
            .drug_type[self.agent.drug_type]
        )

        # HIV
        if (
            pop.pop_random.random() < agent_params.hiv.init
            and time >= pop.params.hiv.start_time
        ):
            self.active = True

            # if HIV, when did the agent convert? Random sample
            self.time = utils.safe_random_int(
                time - self.agent.location.params.hiv.max_init_time,
                time,
                pop.pop_random,
            )

            if pop.pop_random.random() < agent_params.hiv.aids.init:
                self.aids = True

            if pop.pop_random.random() < agent_params.hiv.dx.init:
                self.dx = True
                # agent was diagnosed at a random time between conversion and now
                self.dx_time = utils.safe_random_int(self.time, time, pop.pop_random)

            # add agent to class
            self.add_agent(self.agent)

    def update_agent(self, model: "model.TITAN"):
        """
        Update the agent for this exposure for a time step.  Called once per time step in `TITAN.update_all_agents`. Agent level updates are done after population level updates.   Called on only exposures that are enabled per the params.

        If the agent is hiv+ and the model time is past the hiv start_time, determine if the agent becomes diagnosed if not yet diagnosed, and if the agent has progressed to aids.

        args:
            model: the instance of TITAN currently being run
        """
        if self.active and model.time >= model.params.hiv.start_time:
            if not self.dx:
                test_prob = (
                    self.agent.location.params.demographics[self.agent.race]
                    .sex_type[self.agent.sex_type]
                    .drug_type[self.agent.drug_type]
                    .hiv.dx.prob
                )

                # Rescale based on calibration param
                test_prob *= model.calibration.test_frequency

                if model.run_random.random() < test_prob:
                    self.diagnose(model)

            self.progress_to_aids(model)

    @classmethod
    def add_agent(cls, agent: "agent.Agent"):
        """
        Add an agent to the class (not instance).  This can be useful if tracking population level statistics or groups, such as counts or newly active agents.

        Add the agent to the `agents` set and if the agent is diagnosed, updated the `dx_counts`

        args:
            agent: the agent to add to the class attributes
        """
        cls.agents.add(agent)

        if agent.hiv.dx:  # type: ignore[attr-defined]
            cls.dx_counts[agent.race][agent.sex_type] += 1

    @classmethod
    def remove_agent(cls, agent: "agent.Agent"):
        """
        Remove an agent from the class (not instance).  This can be useful if tracking population level statistics or groups, such as counts.

        Remove the agent from the `agents` set and decrement the `dx_counts` if the agent was diagnosed.

        args:
            agent: the agent to remove from the class attributes
        """
        cls.agents.remove(agent)

        if agent.hiv.dx:  # type: ignore[attr-defined]
            cls.dx_counts[agent.race][agent.sex_type] -= 1

    def set_stats(self, stats: Dict[str, int], time: int):
        if self.active:
            stats["hiv"] += 1
            if self.time == time:
                stats["hiv_new"] += 1
            if self.aids:
                stats["hiv_aids"] += 1
            if self.dx:
                stats["hiv_dx"] += 1
                if self.dx_time == time:
                    stats["hiv_dx_new"] += 1

    @staticmethod
    def expose(
        model: "model.TITAN",
        interaction: str,
        rel: "agent.Relationship",
        num_acts: int,
    ):
        """
        Expose a relationship to the exposure for a number of acts of a specific interaction type.  Typically, this is determining if the exposure can cause conversion/change in one of the agents, then if so determining the probability of that and then converting the succeptible agent.

        For hiv, one agent must be active and the other not for an exposure to cause conversion.

        args:
            model: The running model
            interaction: The type of interaction (e.g. sex, injection)
            rel: The relationship where the interaction is occuring
            num_acts: The number of acts of that interaction
        """
        # Agent 1 is HIV+, Agent 2 is not, Agent 2 is succept
        if rel.agent1.hiv.active and not rel.agent2.hiv.active:  # type: ignore[attr-defined]
            agent = rel.agent1
            partner = rel.agent2
        # If Agent 2 is HIV and Agent 1 is not, Agent 1 is succept
        elif not rel.agent1.hiv.active and rel.agent2.hiv.active:  # type: ignore[attr-defined]
            agent = rel.agent2
            partner = rel.agent1
        else:  # neither agent is HIV or both are
            return

        p = agent.hiv.get_transmission_probability(  # type: ignore[attr-defined]
            model, interaction, partner, num_acts
        )

        if model.run_random.random() < p:
            # if agent HIV+ partner becomes HIV+
            partner.hiv.convert(model)  # type: ignore[attr-defined]

    def get_transmission_probability(
        self,
        model: "model.TITAN",
        interaction: str,
        partner: "agent.Agent",
        num_acts: int,
    ) -> float:
        """
        Determines the probability of an hiv transmission event from agent to partner based on
            interaction type and numer of acts. For sex acts, transmission probability is a
            function of the acquisition probability of the HIV- agent's sex role
            and the HIV+ agent's haart adherence, acute status, and dx risk reduction

        args:
            model: The running model
            interaction : "injection" or "sex"
            partner: HIV- Agent
            num_acts: The number of exposure interactions the agents had this time step

        returns:
            probability of transmission from agent to partner
        """
        # if this isn't an interaction where hiv can transmit, return 0% prob
        if interaction not in ("injection", "sex"):
            return 0.0

        # Logic for if needle or sex type interaction
        p: float

        # get baseline probabilities
        if interaction == "injection":
            p = model.params.partnership.injection.transmission.base
        elif interaction == "sex":
            agent_sex_role = self.agent.sex_role
            partner_sex_role = partner.sex_role

            # get partner's sex role during acts
            if partner_sex_role == "versatile":  # versatile partner takes
                # "opposite" position of agent
                if agent_sex_role == "insertive":
                    partner_sex_role = "receptive"
                elif agent_sex_role == "receptive":
                    partner_sex_role = "insertive"
                else:
                    partner_sex_role = "versatile"  # if both versatile, can switch
                    # between receptive and insertive by act

            # get probability of sex acquisition given HIV- partner's position
            p = partner.location.params.partnership.sex.acquisition[partner.sex_type][
                partner_sex_role
            ]

        # feature specific risk adjustment
        for feature in model.features:
            agent_feature = getattr(self.agent, feature.name)
            p *= agent_feature.get_transmission_risk_multiplier(self.time, interaction)

            partner_feature = getattr(partner, feature.name)
            p *= partner_feature.get_acquisition_risk_multiplier(self.time, interaction)

        # Scaling parameter for acute HIV infections
        if self.get_acute_status(model.time):
            p *= self.agent.location.params.hiv.acute.infectivity

        # Scaling parameter for positively identified HIV agents
        if self.dx:
            p *= 1 - self.agent.location.params.hiv.dx.risk_reduction[interaction]

        # Racial calibration parameter to attain proper race incidence disparity
        p *= partner.location.params.demographics[partner.race].hiv.transmission

        # Scaling parameter for per act transmission.
        p *= model.calibration.acquisition

        return utils.total_probability(p, num_acts)

    def convert(self, model: "model.TITAN"):
        """
        Agent becomes HIV agent. Update all appropriate attributes, sets and dictionaries.

        args:
            model: The model being run
        """
        if not self.active:
            self.active = True
            self.time = model.time
            self.agent.vaccine.active = False  # type: ignore[attr-defined]
            self.add_agent(self.agent)

        if self.agent.prep.active:  # type: ignore[attr-defined]
            self.agent.prep.progress(model, force=True)  # type: ignore[attr-defined]

    def diagnose(self, model: "model.TITAN"):
        """
        Mark the agent as diagnosed.

        args:
             model: the running model
        """
        self.dx = True
        self.dx_time = model.time
        self.add_agent(self.agent)

    # ============================ HELPER METHODS ==============================

    def get_acute_status(self, time: int) -> bool:
        """
        Get acute status of agent at time

        args:
            time: The current time step

        returns:
            whether an agent is acute
        """
        if self.active and self.time is not None:
            hiv_duration = time - self.time

            if self.agent.location.params.hiv.acute.duration >= hiv_duration >= 0:
                return True

        return False

    def progress_to_aids(self, model: "model.TITAN"):
        """
        Model the progression of HIV agents to AIDS agents

        args:
             model: the running model
        """
        aids_params = self.agent.location.params.hiv.aids
        p = self.agent.haart.aids_scale()  # type: ignore[attr-defined]

        if model.run_random.random() < p * aids_params.prob:
            self.aids = True



import random
from typing import Dict, List, Optional
from copy import copy
import os
import logging

import numpy as np  # type: ignore
import nanoid  # type: ignore

from . import agent as ag
from . import output as ao
from . import probabilities as prob
from .parse_params import ObjMap
from . import exposures, features, interactions, population, utils


class TITAN:
    def __repr__(self):
        res = "\n"
        res += f"Seed: {self.run_seed}\n"
        res += f"Npop: {self.params.model.num_pop}\n"
        res += f"Time: {self.params.model.time.num_steps}\n"

        return res

    def __init__(
        self,
        params: ObjMap,
        pop: Optional["population.Population"] = None,
    ):
        """
        This is the core class used to simulate the spread of exposures through a relationship based network.

        args:
            params: the parameter object for this model
            pop: an initialized population to run the model on
        """
        self.id = nanoid.generate(size=8)

        self.params = params
        # pre-fetch commonly used param sub-sets for performance
        self.calibration = params.calibration

        utils.set_up_logging(params)

        logging.info(f"Model ID: {self.id}")
        logging.info("=== Begin Initialization Protocol ===\n")

        if pop is None:
            logging.info("  Generating new population")
            self.pop = population.Population(params)
        else:
            logging.info("  Using provided population")
            self.pop = pop

        self.time = -1 * self.params.model.time.burn_steps  # burn is negative time

        self.features = [
            feature
            for feature in features.BaseFeature.__subclasses__()
            if self.params.features[feature.name]
        ]

        # set up the in-scope exposures
        self.exposures = [
            exposure
            for exposure in exposures.BaseExposure.__subclasses__()
            if self.params.exposures[exposure.name]
        ]

        self.interactions = {
            interaction.name: interaction
            for interaction in interactions.BaseInteraction.__subclasses__()
        }

        # Set seed format. 0: pure random, else: fixed value
        self.run_seed = utils.get_check_rand_int(params.model.seed.run)
        logging.info(f"  Run seed was set to: {self.run_seed}")
        self.run_random = random.Random(self.run_seed)
        self.np_random = np.random.default_rng(self.run_seed)
        random.seed(self.run_seed)
        logging.info(("  FIRST RANDOM CALL {}".format(random.randint(0, 100))))

        logging.info("  Resetting death count")
        self.deaths: List["ag.Agent"] = []  # Number of death

        logging.info("\n=== Initialization Protocol Finished ===")

    def print_stats(self, stat: Dict[str, Dict[str, int]], outdir: str):
        """
        Create/update all of the reports defined in the params
        """
        for report in self.params.outputs.reports:
            printer = getattr(ao, report)
            printer(
                self.id,
                self.time,
                self.run_seed,
                self.pop.pop_seed,
                stat,
                self.params,
                outdir,
            )

        # network-based reports
        if (
            self.time % self.params.outputs.print_frequency == 0
            and self.params.model.network.enable
        ):
            network_outdir = os.path.join(outdir, "network")
            if self.params.outputs.network.calc_component_stats:
                ao.print_components(
                    self.id,
                    self.time,
                    self.run_seed,
                    self.pop.pop_seed,
                    self.pop.connected_components(),
                    network_outdir,
                )

            if self.params.outputs.network.calc_network_stats:
                ao.write_network_stats(
                    self.pop.graph, network_outdir, self.id, self.time
                )

            if self.params.outputs.network.edge_list:
                ao.write_graph_edgelist(
                    self.pop.graph, network_outdir, self.id, self.time
                )

    def reset_trackers(self):
        self.deaths = []

    def run(self, outdir: str):
        """
        Runs the model for the number of time steps defined in params, at each time step does:

        1. Increments time
        2. Takes one step
        3. Resets trackers

        args:
            outdir: path to directory where results should be saved
        """
        # make sure initial state of things get printed
        stats = ao.get_stats(
            self.pop.all_agents,
            self.deaths,
            self.params,
            self.exposures,
            self.features,
            self.time,
        )
        self.print_stats(stats, outdir)

        if self.params.model.time.burn_steps > 0:
            logging.info("  ===! Start Burn Loop !===")

        # time starts at negative burn steps, model run starts at t = 1
        while self.time < self.params.model.time.num_steps:
            if self.time == 0:
                if self.params.model.time.burn_steps > 0:
                    logging.info("  ===! Burn Loop Complete !===")
                logging.info("  ===! Start Main Loop !===")

            self.time += 1
            self.step(outdir)
            self.reset_trackers()

        logging.info("  ===! Main Loop Complete !===")

    def step(self, outdir: str):
        """
        A single time step in the model:

        1. Perform timeline_scaling updates to params if needed
        2. Update all agents
        3. Write/update reports with this timestep's data

        args:
            outdir: path to directory where reports should be saved
        """
        logging.info(
            f"\n                                                  .: TIME {self.time}"
        )
        logging.info(
            "  STARTING HIV count:{}  Total Incarcerated:{}  HR+:{}  "
            "PrEP:{}".format(
                len(exposures.HIV.agents),
                sum([1 for a in self.pop.all_agents if a.incar.active]),  # type: ignore[attr-defined]
                sum([1 for a in self.pop.all_agents if a.high_risk.active]),  # type: ignore[attr-defined]
                sum([1 for a in self.pop.all_agents if a.prep.active]),  # type: ignore[attr-defined]
            )
        )

        self.timeline_scaling()

        self.update_all_agents()

        stats = ao.get_stats(
            self.pop.all_agents,
            self.deaths,
            self.params,
            self.exposures,
            self.features,
            self.time,
        )
        self.print_stats(stats, outdir)

        logging.info(f"Number of relationships: {len(self.pop.relationships)}")
        self.pop.all_agents.print_subsets(logging.info)

    def update_all_agents(self):
        """
        The core of the model.  For a time step, update all of the agents and relationships:


        1. End relationships with no remaining duration
        2. Agent death/replacement
        3. Update partner assignments (create new relationships as needed)
        4. Create an agent zero (if enabled and the time is right)
        5. Agents in relationships interact
        6. Update features at the population level
        7. Update each agent's status for:
            * age
            * all exposures
            * all features (agent level)
        """
        # If static network, ignore relationship progression
        if not self.params.features.static_network:
            for rel in copy(self.pop.relationships):
                if rel.progress():
                    self.pop.remove_relationship(rel)

        if self.params.features.die_and_replace:
            self.die_and_replace()

        if not self.params.features.static_network:
            self.pop.update_partner_assignments(t=self.time)

        # If agent zero enabled, create agent zero at the beginning of main loop.
        if (
            self.time == self.params.agent_zero.start_time
            and self.params.features.agent_zero
        ):
            self.make_agent_zero()

        for rel in self.pop.relationships:
            self.agents_interact(rel)

        for feature in self.features:
            feature.update_pop(self)

        for agent in self.pop.all_agents:
            self.update_agent(agent)

    def update_agent(self, agent):
        """
        Update an agent at the given model timestep.

        Update the agent's status for:
            * age
            * all exposures
            * all features (agent level)
        """
        # happy birthday agents!
        if self.time > 0 and (self.time % self.params.model.time.steps_per_year) == 0:
            agent.age += 1

        for exposure in self.exposures:
            agent_feature = getattr(agent, exposure.name)
            agent_feature.update_agent(self)

        for feature in self.features:
            agent_feature = getattr(agent, feature.name)
            agent_feature.update_agent(self)

    def make_agent_zero(self):
        """
        Identify an agent as agent zero and HIV convert them
        """
        bonds = [  # Find what bond_types have the allowed interaction
            bond
            for bond, act_type in self.params.classes.bond_types.items()
            if self.params.agent_zero.interaction_type in act_type.acts_allowed
        ]
        max_partners = 0
        max_agent = None
        zero_eligible = []
        for agent in self.pop.all_agents:
            num_partners = agent.get_num_partners(bond_types=bonds)
            if num_partners >= self.params.agent_zero.num_partners:
                zero_eligible.append(agent)
            if num_partners > max_partners:
                max_partners = num_partners
                max_agent = agent

        agent_zero = utils.safe_random_choice(zero_eligible, self.run_random)
        if agent_zero:  # if eligible agent, make agent 0
            zero_attr = getattr(agent_zero, self.params.agent_zero.exposure)
            zero_attr.convert(self)
        elif self.params.agent_zero.fallback and max_agent is not None:
            zero_attr = getattr(max_agent, self.params.agent_zero.exposure)
            zero_attr.convert(self)
        else:
            raise ValueError("No agent zero!")

    def timeline_scaling(self):
        """
        Scale/un-scale any params with timeline_scaling definitions per their
        definition.  Applied to all parameters (main model, and location specific).
        """
        if not self.params.features.timeline_scaling:
            return None

        # gather all of the param objectss to be scaled
        params_set = {self.params}
        for location in self.pop.geography.locations.values():
            params_set.add(location.params)

        # iterate over each param and update the values if the time is right
        for params in params_set:
            for defn in params.timeline_scaling.timeline.values():
                param = defn.parameter
                if param != "ts_default":
                    if defn.start_time == self.time:
                        logging.info(f"timeline scaling - {param}")
                        utils.scale_param(params, param, defn.scalar)
                    elif defn.stop_time == self.time:
                        logging.info(f"timeline un-scaling - {param}")
                        utils.scale_param(params, param, 1 / defn.scalar)

    def agents_interact(self, rel: "ag.Relationship"):
        """
        Let an agent interact with a partner.

        Based on the interaction types of the relationship, interact in the following ways:

        * Peer Change Agent
        * Injection
        * Sex

        args:
            rel : The relationship that the agents interact in
        """
        interaction_types = self.params.classes.bond_types[rel.bond_type].acts_allowed
        # If either agent is incarcerated, skip their interaction
        if rel.agent1.incar.active or rel.agent2.incar.active:  # type: ignore[attr-defined]
            return

        for interaction_type in interaction_types:
            interaction = self.interactions[interaction_type]
            interaction.interact(self, rel)

    def die_and_replace(self):
        """
        Let agents die and replace the dead agent with a new agent randomly.
        """
        # die stage
        for agent in self.pop.all_agents:
            # agent incarcerated, don't evaluate for death
            if agent.incar.active:
                continue

            # death rate per 1 person-month
            p = (
                prob.get_death_rate(
                    agent.hiv.active,
                    agent.hiv.aids,
                    agent.drug_type,
                    agent.sex_type,
                    agent.haart.adherent,
                    agent.race,
                    agent.location,
                    self.params.model.time.steps_per_year,
                )
                * self.calibration.mortality
            )

            if self.run_random.random() < p:
                self.deaths.append(agent)

                # End all existing relationships
                for rel in copy(agent.relationships):
                    rel.progress(force=True)
                    self.pop.remove_relationship(rel)

        # replace stage
        for agent in self.deaths:
            # mark agent component as -1 (no componenet)
            agent.component = "-1"

            # Remove agent from agent class and sub-sets
            self.pop.remove_agent(agent)

            new_agent = self.pop.create_agent(
                agent.location, agent.race, self.time, agent.sex_type, agent.drug_type
            )
            self.pop.add_agent(new_agent)



#!/usr/bin/env python
# encoding: utf-8

import random
from collections import deque
from copy import copy
from math import ceil
from typing import List, Dict, Set, Optional, Tuple
import logging

import numpy as np  # type: ignore
import networkx as nx  # type: ignore
import nanoid  # type: ignore

from . import parse_params
from . import agent as ag
from . import location
from . import partnering
from . import utils
from . import features
from . import exposures


class Population:
    def __init__(self, params: "parse_params.ObjMap", id: Optional[str] = None):
        """
        Initialize Population object.

        args:
            params : Model parameters
            id: 8 character identifier for a model
        """
        if id is None:
            self.id = nanoid.generate(size=8)
        else:
            self.id = id

        utils.set_up_logging(params)

        logging.info(f"Population ID: {self.id}")

        self.pop_seed = utils.get_check_rand_int(params.model.seed.ppl)

        # Init RNG for population creation to pop_seed
        self.pop_random = random.Random(self.pop_seed)
        self.np_random = np.random.default_rng(self.pop_seed)

        self.enable_graph = params.model.network.enable
        self.components: List = []

        if self.enable_graph:
            self.graph = nx.Graph()
        else:
            self.graph = None

        self.params = params

        # set up the in-scope exposures
        self.exposures = [
            exposure
            for exposure in exposures.BaseExposure.__subclasses__()
            if self.params.exposures[exposure.name]
        ]
        # initialize the class level items
        for exposure in self.exposures:
            exposure.init_class(params)

        # set up the in-scope features
        self.features = [
            feature
            for feature in features.BaseFeature.__subclasses__()
            if self.params.features[feature.name]
        ]
        # initialize the class level items
        for feature in self.features:
            feature.init_class(params)

        # set up the population's locations and edges
        self.geography = location.Geography(params)

        # All agent set list
        self.all_agents = ag.AgentSet("AllAgents")

        # pwid agents (performance for partnering)
        self.pwid_agents = ag.AgentSet("PWID", parent=self.all_agents)

        # agents who can take on a partner
        self.partnerable_agents: Dict[str, Set["ag.Agent"]] = {}
        for bond_type in self.params.classes.bond_types.keys():
            self.partnerable_agents[bond_type] = set()

        # who can sleep with whom
        self.sex_partners: Dict[str, Set["ag.Agent"]] = {}
        for sex_type in self.params.classes.sex_types.keys():
            self.sex_partners[sex_type] = set()

        self.relationships: Set["ag.Relationship"] = set()

        # find average partnership durations
        self.mean_rel_duration: Dict[str, Dict] = partnering.get_mean_rel_duration(
            self.params
        )

        logging.info("  Creating agents")
        # for each location in the population, create agents per that location's demographics
        init_time = -1 * self.params.model.time.burn_steps
        for loc in self.geography.locations.values():
            for race in params.classes.races:
                for i in range(
                    round(
                        params.model.num_pop
                        * loc.ppl
                        * loc.params.demographics[race].ppl
                    )
                ):
                    if self.all_agents.num_members() >= self.params.model.num_pop:
                        logging.warning(
                            "WARNING: not adding agent to population - too many agents"
                        )
                        break
                    agent = self.create_agent(loc, race, init_time)
                    self.add_agent(agent)

        # initialize relationships
        logging.info("  Creating Relationships")
        self.update_partner_assignments(0)

    def create_agent(
        self,
        loc: "location.Location",
        race: str,
        time: int,
        sex_type: Optional[str] = None,
        drug_type: Optional[str] = None,
    ) -> "ag.Agent":
        """
        Create a new agent with randomly assigned attributes according to population
        demographics [params.demographics]

        args:
            loc: location the agent will live in
            race : race of the new agent
            time: current time step of the model
            sex_type : sex_type of the new agent

        returns:
             a new agent
        """
        if sex_type is None:
            sex_type = utils.safe_random_choice(
                loc.pop_weights[race]["values"],
                self.pop_random,
                weights=loc.pop_weights[race]["weights"],
            )
            # no choice available
            if sex_type is None:
                raise ValueError("Agent must have sex type")

        # Determine drugtype
        if drug_type is None:
            drug_type = utils.safe_random_choice(
                loc.drug_weights[race][sex_type]["values"],
                self.pop_random,
                weights=loc.drug_weights[race][sex_type]["weights"],
            )
            # no choice available
            if drug_type is None:
                raise ValueError("Agent must have drug type")

        age, age_bin = self.get_age(loc, race)

        agent = ag.Agent(sex_type, age, race, drug_type, loc)
        agent.age_bin = age_bin

        sex_role = utils.safe_random_choice(
            loc.role_weights[race][sex_type]["values"],
            self.pop_random,
            weights=loc.role_weights[race][sex_type]["weights"],
        )
        if sex_role is None:
            raise ValueError("Agent must have sex role")
        else:
            agent.sex_role = sex_role

        agent_params = (
            agent.location.params.demographics[race]
            .sex_type[sex_type]
            .drug_type[drug_type]
        )

        for exposure in self.exposures:
            agent_feature = getattr(agent, exposure.name)
            agent_feature.init_agent(self, time)

        for bond, bond_def in loc.params.classes.bond_types.items():
            agent.partners[bond] = set()
            dist_info = agent_params.num_partners[bond]
            agent.mean_num_partners[bond] = ceil(
                utils.safe_dist(dist_info, self.np_random)
                * utils.safe_divide(
                    agent.location.params.calibration.sex.partner,
                    self.mean_rel_duration[bond][race],
                )
            )
            # so not zero if added mid-year
            agent.target_partners[bond] = agent.mean_num_partners[bond]
            if "injection" in bond_def.acts_allowed:
                assert agent.drug_type == "Inj" or agent.mean_num_partners[bond] == 0

            if agent.target_partners[bond] > 0:
                self.partnerable_agents[bond].add(agent)

        for feature in self.features:
            agent_feature = getattr(agent, feature.name)
            agent_feature.init_agent(self, time)

        return agent

    def add_agent(self, agent: "ag.Agent"):
        """
        Adds an agent to the population

        args:
            agent : The agent to be added
        """
        # Add to all agent set
        self.all_agents.add_agent(agent)

        if agent.drug_type == "Inj":
            self.pwid_agents.add_agent(agent)

        # who can sleep with this agent
        for sex_type in self.params.classes.sex_types[agent.sex_type].sleeps_with:
            self.sex_partners[sex_type].add(agent)

        if self.enable_graph:
            self.graph.add_node(agent)

    def add_relationship(self, rel: "ag.Relationship"):
        """
        Add a new relationship to the population.

        args:
            rel : The Relationship to be added
        """
        self.relationships.add(rel)

        if self.enable_graph:
            self.graph.add_edge(rel.agent1, rel.agent2, type=rel.bond_type)

    def remove_agent(self, agent: "ag.Agent"):
        """
        Remove an agent from the population.

        args:
            agent : Agent to remove
        """
        self.all_agents.remove_agent(agent)

        for partner_type in self.sex_partners:
            if agent in self.sex_partners[partner_type]:
                self.sex_partners[partner_type].remove(agent)

        for exposure in self.exposures:
            agent_attr = getattr(agent, exposure.name)
            if agent_attr.active:
                exposure.remove_agent(agent)

        for feature in self.features:
            agent_attr = getattr(agent, feature.name)
            if agent_attr.active:
                feature.remove_agent(agent)

        if self.enable_graph:
            self.graph.remove_node(agent)

        for bond in self.partnerable_agents.values():
            if agent in bond:
                bond.remove(agent)

    def remove_relationship(self, rel: "ag.Relationship"):
        """
        Remove a relationship from the population.

        args:
            rel : Relationship to remove
        """
        self.relationships.remove(rel)

        # without this relationship, are agents partnerable again?
        self.update_partnerability(rel.agent1)
        self.update_partnerability(rel.agent2)

        if self.enable_graph:
            self.graph.remove_edge(rel.agent1, rel.agent2)

    def get_age(self, loc: "location.Location", race: str) -> Tuple[int, int]:
        """
        Given the population characteristics, get a random age to assign to an agent given the race of that agent

        args:
            race : race of the agent whose age is being generated

        returns:
            age and the bin the age came from
        """
        bins = loc.params.demographics[race].age
        i = utils.get_independent_bin(self.pop_random, bins)
        age = self.pop_random.randrange(bins[i].min, bins[i].max)
        return age, i

    def update_agent_partners(
        self, agent: "ag.Agent", bond_type: str, components: List
    ) -> bool:
        """
        Finds and bonds new partner. Creates relationship object for partnership,
            calcs partnership duration, adds it to the population, and adds to networkX graph if self.enable_graph
            is set True.

        args:
            agent: Agent that is seeking a new partner
            bond_type: What type of bond the agent is seeking to make

        returns:
            True if no match was found for agent (used for retries)
        """
        partnerable_agents = self.partnerable_agents[bond_type]
        if (
            self.pop_random.random()
            < self.params.partnership.network.same_component.prob
            and agent.has_partners()
        ):
            # find agent's component
            agent_component: Set["ag.Agent"] = set()
            for comp in components:
                if agent in comp:
                    agent_component = comp
                    break

            partnerable_agents = partnerable_agents & agent_component

        partner = partnering.select_partner(
            agent,
            partnerable_agents,
            self.sex_partners,
            self.pwid_agents,
            self.params,
            self.pop_random,
            bond_type,
        )
        no_match = True

        if partner:
            race = utils.safe_random_choice([agent.race, partner.race], self.pop_random)
            duration = partnering.get_partnership_duration(
                agent.location.params, self.np_random, bond_type, race
            )
            relationship = ag.Relationship(
                agent, partner, duration, bond_type=bond_type
            )
            self.add_relationship(relationship)
            # can partner still partner?
            if len(partner.partners[bond_type]) > (
                partner.target_partners[bond_type]
                * self.params.calibration.partnership.buffer
            ):
                self.partnerable_agents[bond_type].remove(partner)
            no_match = False
        return no_match

    def update_partner_assignments(self, t: int):
        """
        Determines which agents will seek new partners from All_agentSet.
            Calls update_agent_partners for any agents that desire partners.

        args:
            t: current time step of the model
        """
        # update agent targets annually
        if t % self.params.model.time.steps_per_year == 0:
            self.update_partner_targets()

        if self.enable_graph:
            network_components = [set(g.nodes()) for g in self.components]
        else:
            network_components = []

        # Now create partnerships until available partnerships are out
        for bond in self.params.classes.bond_types:
            eligible_agents = deque(
                [
                    a
                    for a in self.all_agents
                    if len(a.partners[bond]) < a.target_partners[bond]
                ]
            )
            attempts = {a: 0 for a in eligible_agents}

            while eligible_agents:
                agent = eligible_agents.popleft()
                if len(agent.partners[bond]) < agent.target_partners[bond]:

                    # no match
                    if self.update_agent_partners(agent, bond, network_components):
                        attempts[agent] += 1

                    # add agent back to eligible pool
                    if (
                        len(agent.partners[bond]) < agent.target_partners[bond]
                        and attempts[agent]
                        < self.params.calibration.partnership.break_point
                    ):
                        eligible_agents.append(agent)

        if self.enable_graph:
            self.trim_graph()

        self.update_agent_components()

    def update_partner_targets(self):
        """
        Update the target number of partners for each agent and bond type
        """
        for a in self.all_agents:
            for bond in self.params.classes.bond_types:
                a.target_partners[bond] = utils.poisson(
                    a.mean_num_partners[bond], self.np_random
                )
            self.update_partnerability(a)

    def update_partnerability(self, a):
        """
        Update whether each agent in the population is currently able to form new relationships for each bond type
        """
        for bond in self.params.classes.bond_types.keys():
            if a in self.partnerable_agents[bond]:
                if len(a.partners[bond]) > (
                    a.target_partners[bond] * self.params.calibration.partnership.buffer
                ):
                    self.partnerable_agents[bond].remove(a)
            elif len(a.partners[bond]) < (
                a.target_partners[bond] * self.params.calibration.partnership.buffer
            ):
                self.partnerable_agents[bond].add(a)

    def update_agent_components(self):
        """
        Update the component IDs associated with each agent based on the current state of the graph
        """
        if self.enable_graph:
            self.components = utils.connected_components(self.graph)
            for id, component in enumerate(self.components):
                for agent in component.nodes:
                    agent.component = str(id)

            self.params.classes.components = list(
                map(str, range(-1, len(self.components)))
            )

    def trim_graph(self):
        """
        Initialize network with graph-based algorithm for relationship
            adding/pruning
        """
        if self.params.model.network.type == "comp_size":

            def trim_component(component, max_size):
                for agent in component.nodes:
                    if (
                        self.pop_random.random()
                        < self.params.calibration.network.trim.prob
                    ):
                        for rel in copy(agent.relationships):
                            if len(agent.relationships) == 1:
                                break  # Make sure that agents stay part of the
                                # network by keeping one bond
                            rel.progress(force=True)
                            self.remove_relationship(rel)

                # recurse on new sub-components
                sub_comps = utils.connected_components(component)
                for sub_comp in sub_comps:
                    if sub_comp.number_of_nodes() > max_size:
                        trim_component(component, max_size)
                    else:
                        break

            components = self.connected_components()
            for comp in components:
                if (
                    comp.number_of_nodes()
                    > self.params.model.network.component_size.max
                ):
                    logging.info("TOO BIG", comp, comp.number_of_nodes())
                    trim_component(comp, self.params.model.network.component_size.max)

        logging.info(f"  Total agents in graph: {self.graph.number_of_nodes()}")

    def connected_components(self) -> List:
        """
        Get connected components in graph (if enabled)

        returns:
            list of connected components
        """
        if self.enable_graph:
            return self.components
        else:
            raise ValueError(
                "Can't get connected_components, population doesn't have graph enabled."
            )



import time as time_mod
from copy import copy
import sys
import os
import shutil
import argparse
import itertools
import json
from multiprocessing import Pool, cpu_count
import csv
import traceback
from typing import List, Optional
import logging

# allow imports to work if running it as a script for development locally
if __name__ == "__main__":
    PACKAGE_PARENT = ".."
    SCRIPT_DIR = os.path.dirname(
        os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__)))
    )
    sys.path.append(os.path.normpath(os.path.join(SCRIPT_DIR, PACKAGE_PARENT)))

from titan.model import TITAN
from titan.population import Population
import titan.population_io as pop_io
from titan.parse_params import create_params
from titan import utils

# how many cores can we use, environment variable returns string
NCORES = int(os.environ.get("SLURM_CPUS_PER_TASK", cpu_count()))

# set up args parsing
parser = argparse.ArgumentParser(description="Run TITAN model")
parser.add_argument(
    "-p", "--params", required=True, help="directory or file with params yaml(s)"
)
parser.add_argument(
    "-S", "--setting", default="custom", help="setting directory to use"
)
parser.add_argument(
    "-n",
    "--nMC",
    type=int,
    nargs="?",
    default=1,
    help="number of monte carlo runs to complete",
)
parser.add_argument(
    "-o", "--outdir", default="results", help="directory name to save results to"
)

parser.add_argument(
    "-e",
    "--error",
    action="store_true",
    help="Error on unused parameters instead of warning",
)

parser.add_argument(
    "--savepop",
    action="store_true",
    help="Save population after creation, but before model run.",
)

parser.add_argument(
    "--poppath",
    type=str,
    default=None,
    help="Path to saved population (directory or .tar.gz file)",
)


def sweep_range(string):
    """
    parsing and checking of param:start:stop:step into dictionary
    """
    error_msg = "Sweep range must have format param:start:stop[:step]"
    parts = string.split(":")
    assert len(parts) in (3, 4), error_msg
    if len(parts) == 4:
        step = parts[3]
    else:
        step = "1"

    try:
        start = int(parts[1])
        stop = int(parts[2])
        step = int(step)
    except ValueError:
        try:
            start = float(parts[1])
            stop = float(parts[2])
            step = float(step)
        except ValueError:
            raise ValueError("start, stop, and step must have same type (int or float)")

    return {"param": parts[0].strip(), "start": start, "stop": stop, "step": step}


parser.add_argument(
    "-w",
    "--sweep",
    nargs="+",
    type=sweep_range,
    default=[],
    help="Optional and repeatable definitions of numeric params to sweep. Expected format is param:start:stop[:step]",
)

parser.add_argument(
    "-W",
    "--sweepfile",
    type=str,
    default=None,
    help="Optional. CSV file with param sweep definitions. Header row must contain param paths, with data rows containing values. If this is passed, any `-w` args will be ignored.",
)

parser.add_argument(
    "-r",
    "--rows",
    type=str,
    default=None,
    help="Optional. Which data rows of sweepfile to use in format start:stop.",
)


parser.add_argument(
    "-F",
    "--force",
    action="store_true",
    help="Run model even if number of sweeps exceeds 100",
)


def drange(start, stop, step):
    """
    Like range, but allows decimal steps
    """
    r = start
    while r < stop:
        yield r
        r += step
        r = round(r, 6)


def setup_sweeps(sweeps):
    """
    Set up sweeps definitions from command line definitions
    """
    sweep_params = [s["param"] for s in sweeps]
    sweep_vals = list(
        itertools.product(
            *[list(drange(s["start"], s["stop"], s["step"])) for s in sweeps]
        )
    )

    defs = []
    for val in sweep_vals:
        defn = {}
        for i, param in enumerate(sweep_params):
            defn[param] = val[i]
        defs.append(defn)

    return defs


def setup_sweeps_file(sweepfile, rows):
    """
    Set up sweeps definitions (params to values) from csv file
    """
    if rows is not None:
        start, stop = (int(item) for item in rows.split(":"))
    else:
        start = 1
        stop = sys.maxsize  # very very big number

    defs = []
    with open(sweepfile, newline="") as f:
        reader = csv.DictReader(f)
        n = 1
        for row in reader:
            if start <= n <= stop:
                defn = {}
                for k, v in row.items():
                    try:
                        val = int(v)
                    except ValueError:
                        try:
                            val = round(float(v), 6)
                        except ValueError:
                            raise ValueError(
                                "sweep values must be numbers (int or float)"
                            )
                    defn[k] = val
                defs.append(defn)
            n += 1

    return defs


def consolidate_files(outdir):
    """
    After running multiple processes, consolidate all of the different processes results
    into the root result directory (outdir)
    """
    for item in os.listdir(outdir):
        subdir = os.path.join(outdir, item)
        if os.path.isdir(subdir) and item not in ("network", "pop"):
            for report in os.listdir(subdir):
                # network folder
                if report == "network":
                    for file in os.listdir(os.path.join(subdir, report)):
                        shutil.move(
                            os.path.join(subdir, report, file),
                            os.path.join(outdir, "network"),
                        )
                elif report == "pop":
                    for file in os.listdir(os.path.join(subdir, report)):
                        shutil.move(
                            os.path.join(subdir, report, file),
                            os.path.join(outdir, "pop"),
                        )
                else:
                    # copy data to existing file
                    if os.path.isfile(os.path.join(outdir, report)):
                        if report == "SweepVals.json":
                            header_skipped = True
                        else:
                            header_skipped = False
                        with open(os.path.join(outdir, report), "a") as tgt, open(
                            os.path.join(subdir, report), "r"
                        ) as src:
                            for line in src:
                                if not header_skipped:
                                    header_skipped = True
                                else:
                                    tgt.write(line)

                    else:
                        shutil.move(os.path.join(subdir, report), os.path.join(outdir))

            # after copying remove directory
            shutil.rmtree(subdir)


def update_sweep_file(run_id, pop_id, defn, outdir):
    """
    Add this run to the sweep file json
    """
    f = open(os.path.join(outdir, "SweepVals.json"), "a")
    res = copy(defn)
    res["run_id"] = run_id
    res["pop_id"] = pop_id
    f.write(json.dumps(res))
    f.write("\n")
    f.close()


def single_run(sweep, outfile_dir, params, save_pop, pop_path):
    """
    A single run of titan.  Dispatched from main using parallel processes.
    """
    utils.set_up_logging(params)

    pid = str(os.getpid())
    pid_outfile_dir = os.path.join(outfile_dir, pid)
    save_pop_dir = (
        os.path.join(pid_outfile_dir, "pop") if save_pop is not None else None
    )
    if not os.path.isdir(pid_outfile_dir):
        os.mkdir(pid_outfile_dir)
        os.mkdir(os.path.join(pid_outfile_dir, "network"))
        if save_pop:
            os.mkdir(save_pop_dir)
        else:
            save_pop_dir = None

    # apply params from sweep for this run
    for param, val in sweep.items():
        utils.override_param(params, param, val, delimiter=".")

    tic = time_mod.time()

    # runs simulations
    if pop_path is None:
        pop = Population(params)
    else:
        pop = pop_io.read(params, pop_path)

    if save_pop_dir is not None:
        pop_io.write(pop, save_pop_dir)
        logging.info(f"Population saved to: {save_pop_dir}")

    try:
        model = TITAN(params, pop=pop)
    except Exception as e:
        raise Exception(f"Model creation failed: {e}")

    try:
        model.run(pid_outfile_dir)
    except Exception as e:
        raise Exception(f"Model run failed for run {model.id}: {e}")

    update_sweep_file(model.id, model.pop.id, sweep, pid_outfile_dir)

    return time_mod.time() - tic


def setup_outdir(outdir, save_pop):
    """
    Set up the results folder - will delete any files already present.
    """
    # delete old results before overwriting with new results
    outfile_dir = os.path.join(os.getcwd(), outdir)
    if os.path.isdir(outfile_dir):
        shutil.rmtree(outfile_dir)
    os.mkdir(outfile_dir)
    os.mkdir(os.path.join(outfile_dir, "network"))
    if save_pop:
        os.mkdir(os.path.join(outfile_dir, "pop"))

    return outfile_dir


def get_sweep_defs(sweepfile, rows, sweeps, num_reps, force):
    """
    Get the sweep definitions from file if available, then from the command line definitions, otherwise return an empty definition to ensure the model runs.
    """
    if sweepfile is not None:
        sweep_defs = setup_sweeps_file(sweepfile, rows)
    elif sweeps:
        sweep_defs = setup_sweeps(sweeps)
    else:
        sweep_defs = [{}]  # make sure runs once

    if len(sweep_defs) > 100 and not force:
        raise ValueError(
            "Sweeping more than 100 models. Set `-F` (force) flag if you really want to do this."
        )

    sweep_defs *= num_reps

    return sweep_defs


def main(
    setting: str,
    params_path: str,
    num_reps: int,
    outdir: str,
    sweeps: List[str],
    force: bool,
    sweepfile: Optional[str] = None,
    rows: Optional[str] = None,
    error_on_unused: bool = False,
    save_pop: bool = False,
    pop_path: Optional[str] = None,
):
    """
    Run TITAN!

    args:
        setting: setting name to use, matches a folder name in `settings/`
        params_path: path to params file or directory
        num_reps: number of time to repeat each sweep
        outdir: directory where results are to be saved
        sweeps: array of strings in param:start:stop:step format
        force: if true, will run even if combination of sweeps results in greater than 100 runs
        sweepfile: path to csv file of sweep definitions
        rows: which rows of the csv to load to create sweeps in start:stop format
        error_on_unused: error if there are parameters that are unused by the model
        save_pop: if true, will save the population to file after creation
        pop_path: path to a population to load instead of creating a new population for each run
    """
    outfile_dir = setup_outdir(outdir, save_pop)

    # generate params - if no setting, set to none
    setting = setting.lower()
    setting_parsed = None if setting == "custom" else setting

    params = create_params(
        setting_parsed,
        params_path,
        outfile_dir,
        error_on_unused=error_on_unused,
    )

    # set up sweeps
    sweep_defs = get_sweep_defs(sweepfile, rows, sweeps, num_reps, force)

    tic = time_mod.time()
    wct = []  # wall clock times

    with Pool(
        processes=NCORES, maxtasksperchild=1
    ) as pool:  # set max tasks/child to prevent processor drift
        results = [
            pool.apply_async(
                single_run, (sweep_def, outfile_dir, params, save_pop, pop_path)
            )
            for sweep_def in sweep_defs
        ]
        while True:
            if all([r.ready() for r in results]):
                break
            else:
                time_mod.sleep(1)

        for r in results:
            try:
                t = r.get()
                wct.append(t)
            except Exception:
                traceback.print_exc()

    toc = time_mod.time() - tic

    consolidate_files(outfile_dir)

    for task, time_t in enumerate(wct):
        print(("wall clock time on for simulation %d: %8.4f seconds" % (task, time_t)))

    def mean(seq):
        return sum(seq) / len(seq)

    print(("\nSUMMARY:\nall tasks - mean: %8.4f seconds" % mean(wct)))
    print(("all tasks - min:  %8.4f seconds" % min(wct)))
    print(("all tasks - max:  %8.4f seconds" % max(wct)))
    print(("all tasks - sum:  %8.4f seconds" % sum(wct)))
    print(f"all tasks - total: {toc} seconds")


def script_init():
    args = parser.parse_args()
    rows = args.rows.strip() if args.rows is not None else None
    sweepfile = args.sweepfile.strip() if args.sweepfile is not None else None
    poppath = args.poppath.strip() if args.poppath is not None else None
    main(
        args.setting.strip(),
        args.params.strip(),
        args.nMC,
        args.outdir.strip(),
        args.sweep,
        args.force,
        sweepfile=sweepfile,
        rows=rows,
        error_on_unused=args.error,
        save_pop=args.savepop,
        pop_path=poppath,
    )


if __name__ == "__main__":
    script_init()



# makes src a package
from .base_feature import *
from .haart import *  # import HAART
from .high_risk import *
from .incar import *
from .external_exposure import *
from .partner_tracing import *
from .prep import *
from .random_trial import *
from .syringe_services import *
from .vaccine import *



from .. import model
from .. import agent


class BaseInteraction:

    name: str = ""
    """Name of interaction in the params file."""

    @classmethod
    def interact(cls, model: "model.TITAN", rel: "agent.Relationship"):
        """
        Given a model and a relation, have the agents in the relationship interact for a time step.

        args:
            model: The running model
            rel: The relationship where interaction is happening
        """
        num_acts = cls.get_num_acts(model, rel)

        if num_acts < 1:
            return

        for exposure in model.exposures:
            if model.time >= model.params[exposure.name].start_time:
                exposure.expose(model, cls.name, rel, num_acts)

    @classmethod
    def get_num_acts(cls, model: "model.TITAN", rel: "agent.Relationship") -> int:
        return 0



from . import base_interaction
from .. import utils
from .. import features
from .. import model
from .. import agent


class Injection(base_interaction.BaseInteraction):

    name = "injection"

    @classmethod
    def get_num_acts(cls, model: "model.TITAN", rel: "agent.Relationship") -> int:
        """
        Simulate random transmission of HIV between two PWID agents through injection.

        args:
            model: The currently running model
            rel: The relationship in which the interaction is happening
        """

        # make sure both agents have Inj drug type, should only be possible for
        # the relationship to have the injection interaction type if both agents PWID
        assert rel.agent1.drug_type == "Inj"
        assert rel.agent2.drug_type == "Inj"

        agent_params = (
            rel.agent1.location.params.demographics[rel.agent1.race]
            .sex_type[rel.agent1.sex_type]
            .injection
        )
        partner_params = (
            rel.agent2.location.params.demographics[rel.agent2.race]
            .sex_type[rel.agent2.sex_type]
            .injection
        )

        mean_num_acts = (
            min(agent_params.num_acts, partner_params.num_acts)
            * model.calibration.injection.act
        )
        share_acts = utils.poisson(mean_num_acts, model.np_random)

        if share_acts < 1:
            return 0

        if (
            rel.agent1.syringe_services.active or rel.agent2.syringe_services.active  # type: ignore[attr-defined]
        ):  # syringe services program risk
            p_unsafe_injection = features.SyringeServices.enrolled_risk
        else:
            p_unsafe_injection = agent_params.unsafe_prob

        # diagnosis risk reduction
        if rel.agent1.hiv.dx or rel.agent1.hiv.dx:  # type: ignore[attr-defined]
            p_unsafe_injection *= 1 - model.params.hiv.dx.risk_reduction.injection

        for n in range(share_acts):
            if model.run_random.random() > p_unsafe_injection:
                share_acts -= 1

        return share_acts
